# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation

from __future__ import annotations
from collections.abc import Iterable

import attrs
from attrs import define, field
import numpy as np
from scipy.interpolate import interp1d

from floris.type_dec import (
    floris_array_converter,
    NDArrayFloat,
    NDArrayFilter,
    NDArrayInt,
    NDArrayObject,
    FromDictMixin,
)
from floris.utilities import cosd
from floris.simulation import BaseClass


def _filter_convert(
    ix_filter: NDArrayFilter | Iterable[int] | None, sample_arg: NDArrayFloat | NDArrayInt
) -> NDArrayFloat | None:
    """This function selects turbine indeces from the given array of turbine properties
    over the simulation's atmospheric conditions (wind directions / wind speeds).
    It converts the ix_filter to a standard format of `np.ndarray`s for filtering
    certain arguments.

    Args:
        ix_filter (NDArrayFilter | Iterable[int] | None): The indices, or truth
            array-like object to use for filtering. None implies that all indeces in the
            sample_arg should be selected.
        sample_arg (NDArrayFloat | NDArrayInt): Any argument that will be filtered, to be used for
            creating the shape. This should be of shape:
            (n wind directions, n wind speeds, n turbines)

    Returns:
        NDArrayFloat | None: Returns an array of a truth or index list if a list is
            passed, a truth array if ix_filter is None, or None if ix_filter is None
            and the `sample_arg` is a single value.
    """
    # Check that the ix_filter is either None or an Iterable. Otherwise,
    # there's nothing we can do with it.
    if not isinstance(ix_filter, Iterable) and ix_filter is not None:
        raise TypeError("Expected ix_filter to be an Iterable or None")

    # Check that the sample_arg is a Numpy array. If it isn't, we
    # can't get its shape.
    if not isinstance(sample_arg, np.ndarray):
        raise TypeError("Expected sample_arg to be a float or integer np.ndarray")

    # At this point, the arguments have this type:
    # ix_filter: Union[Iterable, None]
    # sample_arg: np.ndarray

    # Return all values in the turbine-dimension
    # if the index filter is None
    if ix_filter is None:
        return np.ones(sample_arg.shape[-1], dtype=bool)

    # Finally, we should have an index filter list of type Iterable,
    # so cast it to Numpy array and return
    return np.array(ix_filter)


def power(
    air_density: float,
    velocities: NDArrayFloat,
    yaw_angle: NDArrayFloat,
    pP: float,
    power_interp: NDArrayObject,
    ix_filter: NDArrayInt | Iterable[int] | None = None,
) -> NDArrayFloat:
    """Power produced by a turbine adjusted for yaw and tilt. Value
    given in Watts.

    Args:
        air_density (NDArrayFloat[wd, ws, turbines]): The air density value(s) at each turbine.
        velocities (NDArrayFloat[wd, ws, turbines, grid1, grid2]): The velocity field at a turbine.
        pP (NDArrayFloat[wd, ws, turbines]): The pP value(s) of the cosine exponent relating
            the yaw misalignment angle to power for each turbine.
        power_interp (NDArrayObject[wd, ws, turbines]): The power interpolation function
            for each turbine.
        ix_filter (NDArrayInt, optional): The boolean array, or
            integer indices to filter out before calculation. Defaults to None.

    Returns:
        NDArrayFloat: The power, in Watts, for each turbine after adjusting for yaw and tilt.
    """
    # TODO: Change the order of input arguments to be consistent with the other
    # utility functions - velocities first...
    # Update to power calculation which replaces the fixed pP exponent with
    # an exponent pW, that changes the effective wind speed input to the power
    # calculation, rather than scaling the power.  This better handles power
    # loss to yaw in above rated conditions
    #
    # based on the paper "Optimising yaw control at wind farm level" by
    # Ervin Bossanyi

    # TODO: check this - where is it?
    # P = 1/2 rho A V^3 Cp

    # NOTE: The below has a trivial performance hit for floats being passed (3.4% longer
    # on a meaningless test), but is actually faster when an array is passed through
    # That said, it adds overhead to convert the floats to 1-D arrays, so I don't
    # recommend just converting all values to arrays

    if isinstance(yaw_angle, list):
        yaw_angle = np.array(yaw_angle)

    ix_filter = _filter_convert(ix_filter, yaw_angle)
    if ix_filter is not None:
        velocities = velocities[:, :, ix_filter]
        yaw_angle = yaw_angle[:, :, ix_filter]

    # Compute the yaw effective velocity
    pW = pP / 3.0  # Convert from pP to w
    yaw_effective_velocity = average_velocity(velocities) * cosd(yaw_angle) ** pW
    p = power_interp(yaw_effective_velocity)
    return p * air_density


def Ct(
    velocities: NDArrayFloat,
    yaw_angle: NDArrayFloat,
    fCt: NDArrayObject,
    ix_filter: NDArrayFilter | Iterable[int] | None = None,
) -> NDArrayFloat:

    """Thrust coefficient of a turbine incorporating the yaw angle.
    The value is interpolated from the coefficient of thrust vs
    wind speed table using the rotor swept area average velocity.

    Args:
        velocities (NDArrayFloat[wd, ws, turbines, grid1, grid2]): The velocity field at a turbine.
        yaw_angle (NDArrayFloat[wd, ws, turbines]): The yaw angle for each turbine.
        fCt (NDArrayObject[wd, ws, turbines]): The thrust coefficient for each turbine.
        ix_filter (NDArrayFilter | Iterable[int] | None, optional): The boolean array, or
            integer indices as an iterable of array to filter out before calculation. Defaults to None.

    Returns:
        NDArrayFloat: Coefficient of thrust.
    """

    if isinstance(yaw_angle, list):
        yaw_angle = np.array(yaw_angle)

    if ix_filter is not None:
        ix_filter = _filter_convert(ix_filter, yaw_angle)
        velocities = velocities[:, :, ix_filter]
        yaw_angle = yaw_angle[:, :, ix_filter]

    average_velocities = average_velocity(velocities)
    thrust_coefficient = fCt(average_velocities)
    thrust_coefficient = np.clip(thrust_coefficient, 0.0001, 0.9999)
    effective_thrust = thrust_coefficient * cosd(yaw_angle)
    return effective_thrust


def axial_induction(
    velocities: NDArrayFloat,  # (wind directions, wind speeds, turbines, grid, grid)
    yaw_angle: NDArrayFloat,  # (wind directions, wind speeds, turbines)
    fCt: NDArrayObject,  # (wind directions, wind speeds, turbines)
    ix_filter: NDArrayFilter | Iterable[int] | None = None,
) -> NDArrayFloat:
    """Axial induction factor of the turbine incorporating
    the thrust coefficient and yaw angle.

    Args:
        velocities (NDArrayFloat): The velocity field at each turbine; should be shape:
            (number of turbines, ngrid, ngrid), or (ngrid, ngrid) for a single turbine.
        fCt (np.array): The thrust coefficient function for each
            turbine.
        ix_filter (NDArrayFilter | Iterable[int] | None, optional): The boolean array, or
            integer indices (as an aray or iterable) to filter out before calculation.
            Defaults to None.

    Returns:
        Union[float, NDArrayFloat]: [description]
    """

    if isinstance(yaw_angle, list):
        yaw_angle = np.array(yaw_angle)

    # Get Ct first before modifying any data
    thrust_coefficient = Ct(velocities, yaw_angle, fCt, ix_filter)

    # Then, process the input arguments as needed for this function
    ix_filter = _filter_convert(ix_filter, yaw_angle)
    if ix_filter is not None:
        yaw_angle = yaw_angle[:, :, ix_filter]

    return 0.5 / cosd(yaw_angle) * (1 - np.sqrt(1 - thrust_coefficient * cosd(yaw_angle)))


def average_velocity(velocities: NDArrayFloat, ix_filter: NDArrayFilter | Iterable[int] | None = None) -> NDArrayFloat:
    """This property calculates and returns the cube root of the
    mean cubed velocity in the turbine's rotor swept area (m/s).

    **Note:** The velocity is scaled to an effective velocity by the yaw.

    Args:
        velocities (NDArrayFloat): The velocity field at each turbine; should be shape:
            (number of turbines, ngrid, ngrid), or (ngrid, ngrid) for a single turbine.
        ix_filter (NDArrayFilter | Iterable[int] | None], optional): The boolean array, or
            integer indices (as an iterable or array) to filter out before calculation.
            Defaults to None.

    Returns:
        NDArrayFloat: The average velocity across the rotor(s).
    """
    # Remove all invalid numbers from interpolation
    # data = np.array(self.velocities)[~np.isnan(self.velocities)]

    # The input velocities are expected to be a 5 dimensional array with shape:
    # (# wind directions, # wind speeds, # turbines, grid resolution, grid resolution)

    if ix_filter is not None:
        velocities = velocities[:, :, ix_filter]
    axis = tuple([3 + i for i in range(velocities.ndim - 3)])
    return np.cbrt(np.mean(velocities ** 3, axis=axis))


@define
class PowerThrustTable(FromDictMixin):
    """Helper class to convert the dictionary and list-based inputs to a object of arrays.

    Args:
        power (NDArrayFloat): The power produced at a given windspeed.
        thrust (NDArrayFloat): The thrust at a given windspeed.
        wind_speed (NDArrayFloat): Windspeed values, m/s.

    Raises:
        ValueError: Raised if the power, thrust, and wind_speed are not all 1-d array-like shapes.
        ValueError: Raised if power, thrust, and wind_speed don't have the same number of values.
    """
    power: NDArrayFloat = field(converter=floris_array_converter)
    thrust: NDArrayFloat = field(converter=floris_array_converter)
    wind_speed: NDArrayFloat = field(converter=floris_array_converter)

    def __attrs_post_init__(self) -> None:
        # Validate the power, thrust, and wind speed inputs.

        inputs = (self.power, self.thrust, self.wind_speed)

        if any(el.ndim > 1 for el in inputs):
            raise ValueError("power, thrust, and wind_speed inputs must be 1-D.")

        if len( set( (self.power.size, self.thrust.size, self.wind_speed.size) ) ) > 1:        
            raise ValueError("power, thrust, and wind_speed tables must be the same size.")
        
        # Remove any duplicate wind speed entries
        _, duplicate_filter = np.unique(self.wind_speed, return_index=True)
        object.__setattr__(self, "power", self.power[duplicate_filter])
        object.__setattr__(self, "thrust", self.thrust[duplicate_filter])
        object.__setattr__(self, "wind_speed", self.wind_speed[duplicate_filter])


@define
class Turbine(BaseClass):
    """
    Turbine is a class containing objects pertaining to the individual
    turbines.

    Turbine is a model class representing a particular wind turbine. It
    is largely a container of data and parameters, but also contains
    methods to probe properties for output.

    Parameters:
        rotor_diameter (:py:obj: float): The rotor diameter (m).
        hub_height (:py:obj: float): The hub height (m).
        pP (:py:obj: float): The cosine exponent relating the yaw
            misalignment angle to power.
        pT (:py:obj: float): The cosine exponent relating the rotor
            tilt angle to power.
        generator_efficiency (:py:obj: float): The generator
            efficiency factor used to scale the power production.
        power_thrust_table (PowerThrustTable): A dictionary containing the
            following key-value pairs:

            power (:py:obj: List[float]): The coefficient of power at
                different wind speeds.
            thrust (:py:obj: List[float]): The coefficient of thrust
                at different wind speeds.
            wind_speed (:py:obj: List[float]): The wind speeds for
                which the power and thrust values are provided (m/s).
        ngrid (*int*, optional): The square root of the number
            of points to use on the turbine grid. This number will be
            squared so that the points can be evenly distributed.
            Defaults to 5.
        rloc (:py:obj: float, optional): A value, from 0 to 1, that determines
            the width/height of the grid of points on the rotor as a ratio of
            the rotor radius.
            Defaults to 0.5.
    """

    rotor_diameter: float = field()
    hub_height: float
    pP: float
    pT: float
    TSR: float
    generator_efficiency: float
    power_thrust_table: PowerThrustTable = field(converter=PowerThrustTable.from_dict)

    # rloc: float = float_attrib()  # TODO: goes here or on the Grid?
    # use_points_on_perimeter: bool = bool_attrib()

    # Initialized in the post_init function
    rotor_radius: float = field(init=False)
    rotor_area: float = field(init=False)
    fCp_interp: interp1d = field(init=False)
    fCt_interp: interp1d = field(init=False)
    power_interp: interp1d = field(init=False)

    # For the following parameters, use default values if not user-specified
    # self.rloc = float(input_dictionary["rloc"]) if "rloc" in input_dictionary else 0.5
    # if "use_points_on_perimeter" in input_dictionary:
    #     self.use_points_on_perimeter = bool(input_dictionary["use_points_on_perimeter"])
    # else:
    #     self.use_points_on_perimeter = False

    def __attrs_post_init__(self) -> None:

        # Post-init initialization for the power curve interpolation functions
        wind_speeds = self.power_thrust_table.wind_speed
        self.fCp_interp = interp1d(
            wind_speeds,
            self.power_thrust_table.power,
            fill_value=(0.0, 1.0),
            bounds_error=False,
        )
        inner_power = 0.5 * self.rotor_area * self.fCp_interp(wind_speeds) * self.generator_efficiency * wind_speeds ** 3
        self.power_interp = interp1d(
            wind_speeds,
            inner_power,
            fill_value="extrapolate"
        )

        """
        Given an array of wind speeds, this function returns an array of the interpolated thrust coefficients
        from the power / thrust table used to define the Turbine. The values are bound by the range of the input values.
        Any requested wind speeds outside of the range of input wind speeds are assigned Ct of 0.0001 or 0.9999.
        
        The fill_value arguments sets (upper, lower) bounds for any values outside of the input range
        """
        self.fCt_interp = interp1d(
            wind_speeds,
            self.power_thrust_table.thrust,
            fill_value=(0.0001, 0.9999),
            bounds_error=False,
        )

    @rotor_diameter.validator
    def reset_rotor_diameter_dependencies(self, instance: attrs.Attribute, value: float) -> None:
        """Resets the `rotor_radius` and `rotor_area` attributes."""
        # Temporarily turn off validators to avoid infinite recursion
        with attrs.validators.disabled():
            # Reset the values
            self.rotor_radius = value / 2.0
            self.rotor_area = np.pi * self.rotor_radius ** 2.0

    @rotor_radius.validator
    def reset_rotor_radius(self, instance: attrs.Attribute, value: float) -> None:
        """
        Resets the `rotor_diameter` value to trigger the recalculation of
        `rotor_diameter`, `rotor_radius` and `rotor_area`.
        """
        self.rotor_diameter = value * 2.0

    @rotor_area.validator
    def reset_rotor_area(self, instance: attrs.Attribute, value: float) -> None:
        """
        Resets the `rotor_radius` value to trigger the recalculation of
        `rotor_diameter`, `rotor_radius` and `rotor_area`.
        """
        self.rotor_radius = (value / np.pi) ** 0.5



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation


import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import norm

from ...visualization import visualize_cut_plane


class Yaw:
    """
    Class that performs yaw optimization for a single set of
    inflow conditions. Intended to be used together with an object of the
    :py:class`floris.tools.optimization.optimization.Optimization` class.

    Args:
        fi (:py:class:`floris.tools.floris_interface.FlorisInterface`):
            Interface from FLORIS to the tools package.
        minimum_yaw_angle (float, optional): Minimum constraint on
            yaw. Defaults to None.
        maximum_yaw_angle (float, optional): Maximum constraint on
            yaw. Defaults to None.
        x0 (iterable, optional): The initial yaw conditions.
            Defaults to None. Initializes to the current turbine
            yaw settings.
        include_unc (bool): If True, uncertainty in wind direction
            and/or yaw position is included when determining wind farm power.
            Uncertainty is included by computing the mean wind farm power for
            a distribution of wind direction and yaw position deviations from
            the original wind direction and yaw angles. Defaults to False.
        unc_pmfs (dictionary, optional): A dictionary containing optional
            probability mass functions describing the distribution of wind
            direction and yaw position deviations when wind direction and/or
            yaw position uncertainty is included in the power calculations.
            Contains the following key-value pairs:

            -   **wd_unc**: A numpy array containing wind direction deviations
                from the original wind direction.
            -   **wd_unc_pmf**: A numpy array containing the probability of
                each wind direction deviation in **wd_unc** occuring.
            -   **yaw_unc**: A numpy array containing yaw angle deviations
                from the original yaw angles.
            -   **yaw_unc_pmf**: A numpy array containing the probability of
                each yaw angle deviation in **yaw_unc** occuring.

            Defaults to None, in which case default PMFs are calculated using
            values provided in **unc_options**.
        unc_options (disctionary, optional): A dictionary containing values used
            to create normally-distributed, zero-mean probability mass functions
            describing the distribution of wind direction and yaw position
            deviations when wind direction and/or yaw position uncertainty is
            included. This argument is only used when **unc_pmfs** is None and
            contains the following key-value pairs:

            -   **std_wd**: A float containing the standard deviation of the wind
                    direction deviations from the original wind direction.
            -   **std_yaw**: A float containing the standard deviation of the yaw
                    angle deviations from the original yaw angles.
            -   **pmf_res**: A float containing the resolution in degrees of the
                    wind direction and yaw angle PMFs.
            -   **pdf_cutoff**: A float containing the cumulative distribution
                function value at which the tails of the PMFs are truncated.

            Defaults to None. Initializes to {'std_wd': 4.95, 'std_yaw': 1.75,
            'pmf_res': 1.0, 'pdf_cutoff': 0.995}.
        wdir (float, optional): Wind direction to use for optimization. Defaults
            to None. Initializes to current wind direction in floris.
        wspd (float, optional): Wind speed to use for optimization. Defaults
            to None. Initializes to current wind direction in floris.

    Returns:
        Yaw: An instantiated Yaw object.
    """

    def __init__(
        self,
        fi,
        minimum_yaw_angle=0.0,
        maximum_yaw_angle=25.0,
        x0=None,
        include_unc=False,
        unc_pmfs=None,
        unc_options=None,
        wdir=None,
        wspd=None,
    ):
        """
        Instantiate Yaw object and parameter values.
        """
        self.fi = fi
        self.minimum_yaw_angle = minimum_yaw_angle
        self.maximum_yaw_angle = maximum_yaw_angle

        if x0 is not None:
            self.x0 = x0
        else:
            self.x0 = [
                turbine.yaw_angle
                for turbine in self.fi.floris.farm.turbine_map.turbines
            ]

        self.include_unc = include_unc
        self.unc_pmfs = unc_pmfs
        if self.include_unc & (self.unc_pmfs is None):
            self.unc_pmfs = calc_unc_pmfs(self.unc_pmfs)

        if wdir is not None:
            self.wdir = wdir
        else:
            self.wdir = self.fi.floris.farm.flow_field.wind_direction
        if wspd is not None:
            self.wspd = wspd
        else:
            self.wspd = self.fi.floris.farm.flow_field.wind_speed

        self.fi.reinitialize_flow_field(wind_speed=self.wspd, wind_direction=self.wdir)

    def __str__(self):
        return "yaw"

    ###########################################################################
    # Required private optimization methods
    ###########################################################################

    def reinitialize(self):
        pass

    def obj_func(self, varDict):
        # Parse the variable dictionary
        self.parse_opt_vars(varDict)

        # Reinitialize with wind speed and direction
        self.fi.reinitialize_flow_field(wind_speed=self.wspd, wind_direction=self.wdir)

        # Compute the objective function
        funcs = {}
        funcs["obj"] = -1 * self.fi.get_farm_power_for_yaw_angle(self.yaw) / 1e0

        # Compute constraints, if any are defined for the optimization
        funcs = self.compute_cons(funcs)

        fail = False
        return funcs, fail

    def parse_opt_vars(self, varDict):
        self.yaw = varDict["yaw"]

    def parse_sol_vars(self, sol):
        self.yaw = list(sol.getDVs().values())[0]

    def add_var_group(self, optProb):
        optProb.addVarGroup(
            "yaw",
            self.nturbs,
            type="c",
            lower=self.minimum_yaw_angle,
            upper=self.maximum_yaw_angle,
            value=self.x0,
        )

        return optProb

    def add_con_group(self, optProb):
        # no constraints defined
        return optProb

    def compute_cons(self, funcs):
        # no constraints defined
        return funcs

    ###########################################################################
    # User-defined methods
    ###########################################################################

    def plot_yaw_opt_results(self, sol):
        """
        Method to plot the wind farm with optimal yaw offsets
        """
        yaw = sol.getDVs()["yaw"]

        # Assign yaw angles to turbines and calculate wake
        self.fi.calculate_wake(yaw_angles=yaw)

        # Initialize the horizontal cut
        horizontal_plane = self.fi.calculate_horizontal_plane(x_resolution=400, y_resolution=100)

        # Plot and show
        fig, ax = plt.subplots()
        visualize_cut_plane(horizontal_plane, ax=ax)
        ax.set_title(
            "Optimal Yaw Offsets for U = "
            + str(self.wspd[0])
            + " m/s, Wind Direction = "
            + str(self.wdir[0])
            + "$^\\circ$"
        )

        plt.show()

    def print_power_gain(self, sol):
        """
        Method to print the power gain from wake steering with optimal yaw offsets
        """
        yaw = sol.getDVs()["yaw"]

        self.fi.calculate_wake(yaw_angles=0.0)
        power_baseline = self.fi.get_farm_power()

        self.fi.calculate_wake(yaw_angles=yaw)
        power_opt = self.fi.get_farm_power()

        pct_gain = 100.0 * (power_opt - power_baseline) / power_baseline

        print("==========================================")
        print("Baseline Power = %.1f kW" % (power_baseline / 1e3))
        print("Optimal Power = %.1f kW" % (power_opt / 1e3))
        print("Total Power Gain = %.1f%%" % pct_gain)
        print("==========================================")

    ###########################################################################
    # Properties
    ###########################################################################

    @property
    def nturbs(self):
        """
        This property returns the number of turbines in the FLORIS
        object.

        Returns:
            nturbs (int): The number of turbines in the FLORIS object.
        """
        self._nturbs = len(self.fi.floris.farm.turbines)
        return self._nturbs


def calc_unc_pmfs(unc_options=None):
    """
    Calculates normally-distributed probability mass functions describing the
    distribution of wind direction and yaw position deviations when wind direction
    and/or yaw position uncertainty are included in power calculations.

    Args:
        unc_options (dictionary, optional): A dictionary containing values used
                to create normally-distributed, zero-mean probability mass functions
                describing the distribution of wind direction and yaw position
                deviations when wind direction and/or yaw position uncertainty is
                included. This argument is only used when **unc_pmfs** is None and
                contains the following key-value pairs:

                -   **std_wd**: A float containing the standard deviation of the wind
                        direction deviations from the original wind direction.
                -   **std_yaw**: A float containing the standard deviation of the yaw
                        angle deviations from the original yaw angles.
                -   **pmf_res**: A float containing the resolution in degrees of the
                        wind direction and yaw angle PMFs.
                -   **pdf_cutoff**: A float containing the cumulative distribution
                    function value at which the tails of the PMFs are truncated.

                Defaults to None. Initializes to {'std_wd': 4.95, 'std_yaw': 1.75,
                'pmf_res': 1.0, 'pdf_cutoff': 0.995}.

    Returns:
        [dictionary]: A dictionary containing
                probability mass functions describing the distribution of wind
                direction and yaw position deviations when wind direction and/or
                yaw position uncertainty is included in the power calculations.
                Contains the following key-value pairs:

                -   **wd_unc**: A numpy array containing wind direction deviations
                    from the original wind direction.
                -   **wd_unc_pmf**: A numpy array containing the probability of
                    each wind direction deviation in **wd_unc** occuring.
                -   **yaw_unc**: A numpy array containing yaw angle deviations
                    from the original yaw angles.
                -   **yaw_unc_pmf**: A numpy array containing the probability of
                    each yaw angle deviation in **yaw_unc** occuring.

    """

    if unc_options is None:
        unc_options = {
            "std_wd": 4.95,
            "std_yaw": 1.75,
            "pmf_res": 1.0,
            "pdf_cutoff": 0.995,
        }

    # create normally distributed wd and yaw uncertainty pmfs
    if unc_options["std_wd"] > 0:
        wd_bnd = int(
            np.ceil(
                norm.ppf(unc_options["pdf_cutoff"], scale=unc_options["std_wd"])
                / unc_options["pmf_res"]
            )
        )
        wd_unc = np.linspace(
            -1 * wd_bnd * unc_options["pmf_res"],
            wd_bnd * unc_options["pmf_res"],
            2 * wd_bnd + 1,
        )
        wd_unc_pmf = norm.pdf(wd_unc, scale=unc_options["std_wd"])
        wd_unc_pmf = wd_unc_pmf / np.sum(wd_unc_pmf)  # normalize so sum = 1.0
    else:
        wd_unc = np.zeros(1)
        wd_unc_pmf = np.ones(1)

    if unc_options["std_yaw"] > 0:
        yaw_bnd = int(
            np.ceil(
                norm.ppf(unc_options["pdf_cutoff"], scale=unc_options["std_yaw"])
                / unc_options["pmf_res"]
            )
        )
        yaw_unc = np.linspace(
            -1 * yaw_bnd * unc_options["pmf_res"],
            yaw_bnd * unc_options["pmf_res"],
            2 * yaw_bnd + 1,
        )
        yaw_unc_pmf = norm.pdf(yaw_unc, scale=unc_options["std_yaw"])
        yaw_unc_pmf = yaw_unc_pmf / np.sum(yaw_unc_pmf)  # normalize so sum = 1.0
    else:
        yaw_unc = np.zeros(1)
        yaw_unc_pmf = np.ones(1)

    return {
        "wd_unc": wd_unc,
        "wd_unc_pmf": wd_unc_pmf,
        "yaw_unc": yaw_unc,
        "yaw_unc_pmf": yaw_unc_pmf,
    }



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation


import numpy as np

from ..utilities import wrap_180, wrap_360


def log_law_interpolate(z_test, z_ref, v_ref, roughness=0.03):
    """
    Interpolate wind speed assuming a log-law profile.

    Args:
        z_test (float): height of interest for wind speed estimate.
        z_ref (float): reference height.
        v_ref (float): reference velocity.
        roughness (float, optional): Effective roughness length.
            Defaults to 0.03.

    Returns:
        v_test (np.float): interpolated wind speed at z_test.
    """
    return v_ref * np.log(z_test / roughness) / np.log(z_ref / roughness)


def determine_rews_weights(R, HH, heights_in):
    """
    Weighting for rotor-equivalent wind speed (REWS).

    Args:
        R (float): rotor diameter.
        HH (float): hub height.
        heights_in (iterable): heights of interest.

    Returns:
        weights_return (list): list of weighting values for REWS.
    """
    # Remove any heights not in range of the rotor
    heights = [h for h in heights_in if ((h >= HH - R) and (h <= HH + R))]
    num_heights = len(heights)

    # Determine the zone interfaces
    zone_boundaries = np.zeros(num_heights + 1)
    zone_boundaries[0] = HH - R
    zone_boundaries[-1] = HH + R
    for i in range(1, num_heights):
        zone_boundaries[i] = (heights[i] - heights[i - 1]) / 2.0 + heights[i - 1]
    zone_interfaces = zone_boundaries[1:-1]

    # Next find the central angles for each interace
    h = zone_interfaces - HH
    alpha = np.arcsin(h / R)
    C = np.pi - 2 * alpha
    A = ((R ** 2) / 2) * (C - np.sin(C))
    A = [np.pi * R ** 2] + [a for a in A]
    for i in range(num_heights - 1):
        A[i] = A[i] - A[i + 1]
    weights = A

    # normalize
    weights = weights / np.sum(weights)

    # Now re-pad weights to include heights that were initally cropped
    weight_dict = dict(zip(heights, weights))
    weights_return = [weight_dict.get(h, 0.0) for h in heights_in]

    return weights_return


def rews_from_df(df, columns_in, weights, rews_name, circular=False):
    """
    Estimate the rotor-equivalent wind speed (REWS) from wind speed.

    Args:
        df (pd.DataFrame): DataFrame containing flow information
        columns_in (list): columns to include estimate of REWS.
        weights (iterable): weighting values for REWS.
        rews_name (str): column name for REWS output.
        circular (bool, optional): flag to consider REWS azimuthally.
            Defaults to False.

    Returns:
        df (pd.DataFrame): updated dataframe with REWS column.
    """
    # Ensure numpy array
    weights = np.array(weights)

    # Get the data
    data_matrix = df[columns_in].values

    if not circular:
        df[rews_name] = compute_rews(data_matrix, weights)
    else:
        cos_vals = compute_rews(np.cos(np.deg2rad(data_matrix)), weights)
        sin_vals = compute_rews(np.sin(np.deg2rad(data_matrix)), weights)
        df[rews_name] = wrap_360(np.rad2deg(np.arctan2(sin_vals, cos_vals)))

    return df


def compute_rews(data_matrix, weights):
    """
    Calculation method for REWS from wind speed and weighting values.

    Args:
        data_matrix (np.array): wind speed data
        weights (np.array): weighting values for REWS.

    Returns:
        REWS (float): rotor-equivalent wind speed.
    """
    return np.sum(data_matrix * weights, axis=1)



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation


from floris.simulation.wake_deflection.gauss import GaussVelocityDeflection
from floris.simulation.wake_deflection.jimenez import JimenezVelocityDeflection



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation

import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import minimize

from .layout import LayoutOptimization
from .base_COE import BaseCOE


class LayoutHeightOptimization(LayoutOptimization):
    """
    LayoutHeightOptimization is a subclass of
    :py:class:`~.tools.optimization.scipy.layout.LayoutOptimization` that
    performs layout and turbine height optimization. This optimization method
    aims to minimize Cost of Energy (COE) by changing individual turbine
    locations and all turbine heights across the wind farm. Note that the
    changing turbine height applies to all turbines, i.e. although the turbine
    height is changing, all turbines will be assigned the same turbine height.
    """

    def __init__(
        self,
        fi,
        boundaries,
        height_lims,
        wd,
        ws,
        freq,
        AEP_initial,
        COE_initial,
        plant_kw,
        x0=None,
        bnds=None,
        min_dist=None,
        opt_method="SLSQP",
        opt_options=None,
    ):
        """
        Instantiate LayoutHeightOptimization object with a FlorisInterface
        object and assign parameter values.

        Args:
            fi (:py:class:`~.tools.floris_interface.FlorisInterface`):
                Interface used to interact with the Floris object.
            boundaries (iterable(float, float)): Pairs of x- and y-coordinates
                that represent the boundary's vertices (m).
            height_lims (iterable): A list of the minimum and maximum
                height limits for the optimization (m). Each value only
                needs to be defined once since all the turbine heights
                are the same (ie. [h_min, h_max]).
            wd (np.array): An array of wind directions (deg).
            ws (np.array): An array of wind speeds (m/s).
            freq (np.array): An array of the frequencies of occurance
                correponding to each pair of wind direction and wind speed
                values.
            AEP_initial (float): The initial Annual Energy
                Production used for normalization in the optimization (Wh)
                (TODO: Is Watt-hours the correct unit?).
            COE_initial (float): Initial Cost of Energy used for
                normalization in the optimization ($/kWh).
            plant_kw (float): The rating of the entire wind plant (kW).
            x0 (iterable, optional): The initial turbine locations,
                ordered by x-coordinate and then y-coordiante
                (ie. [x1, x2, ..., xn, y1, y2, ..., yn]), and the
                initial turbine hub height (m). If none are provided, x0
                initializes to the current turbine locations and hub height.
                Defaults to None.
            bnds (iterable, optional): Bounds for the optimization
                variables (TODO: just coordinates, or height too?) (pairs of
                min/max values for each variable (m)). If none are specified,
                they are set to the min. and max. of the boundaries iterable.
                Defaults to None.
            min_dist (float, optional): The minimum distance to be
                maintained between turbines during the optimization (m). If not
                specified, initializes to 2 rotor diameters. Defaults to None.
            opt_method (str, optional): The optimization method used by
                scipy.optimize.minize. Defaults to 'SLSQP'.
            opt_options (dict, optional): Optimization options used by
                scipy.optimize.minize. If none are specified, they are set to
                {'maxiter': 100, 'disp': True, 'iprint': 2, 'ftol': 1e-9}.
                Defaults to None.
        """
        super().__init__(fi, boundaries, wd, ws, freq, AEP_initial)
        self.epsilon = np.finfo(float).eps

        self.COE_model = BaseCOE(self)

        self.reinitialize_opt_height(
            boundaries=boundaries,
            height_lims=height_lims,
            wd=wd,
            ws=ws,
            freq=freq,
            AEP_initial=AEP_initial,
            COE_initial=COE_initial,
            plant_kw=plant_kw,
            x0=x0,
            bnds=bnds,
            min_dist=min_dist,
            opt_method=opt_method,
            opt_options=opt_options,
        )

    # Private methods

    def _fCp_outside(self):
        pass  # for future use

    def _fCt_outside(self):
        pass  # for future use

    def _set_initial_conditions(self):
        self.x0.append(
            self._norm(
                self.fi.floris.farm.turbines[0].hub_height, self.bndh_min, self.bndh_max
            )
        )

    def _set_opt_bounds_height(self):
        self.bnds.append((0.0, 1.0))

    def _optimize(self):
        self.residual_plant = minimize(
            self._COE_layout_height_opt,
            self.x0,
            method=self.opt_method,
            bounds=self.bnds,
            constraints=self.cons,
            options=self.opt_options,
        )

        opt_results = self.residual_plant.x

        return opt_results

    def _COE_layout_height_opt(self, opt_vars):
        locs = self._unnorm(opt_vars[0 : 2 * self.nturbs], self.bndx_min, self.bndx_max)
        height = self._unnorm(opt_vars[-1], self.bndh_min, self.bndh_max)

        self._change_height(height)
        self._change_coordinates(locs)
        AEP_sum = self._AEP_loop_wd()
        COE = self.COE_model.COE(height, AEP_sum)

        return COE / self.COE_initial

    def _change_height(self, height):
        if isinstance(height, float) or isinstance(height, int):
            for turb in self.fi.floris.farm.turbines:
                turb.hub_height = height
        else:
            for k, turb in enumerate(self.fi.floris.farm.turbines):
                turb.hub_height = height[k]

        self.fi.reinitialize_flow_field(
            layout_array=((self.fi.layout_x, self.fi.layout_y))
        )

    # Public methods

    def reinitialize_opt_height(
        self,
        boundaries=None,
        height_lims=None,
        wd=None,
        ws=None,
        freq=None,
        AEP_initial=None,
        COE_initial=None,
        plant_kw=None,
        x0=None,
        bnds=None,
        min_dist=None,
        opt_method=None,
        opt_options=None,
    ):
        """
        This method reinitializes any optimization parameters that are
        specified. Otherwise, the current parameter values are kept.

        Args:
            boundaries (iterable(float, float)): Pairs of x- and y-coordinates
                that represent the boundary's vertices (m).
            height_lims (iterable): A list of the minimum and maximum
                height limits for the optimization (m). Each value only
                needs to be defined once since all the turbine heights
                are the same (ie. [h_min, h_max]). Defaults to None.
            wd (np.array): An array of wind directions (deg). Defaults to None.
            ws (np.array): An array of wind speeds (m/s). Defaults to None.
            freq (np.array): An array of the frequencies of occurance
                correponding to each pair of wind direction and wind speed
                values. Defaults to None.
            AEP_initial (float): The initial Annual Energy
                Production used for normalization in the optimization (Wh).
                Defaults to None.
            COE_initial (float): Initial Cost of Energy used for
                normalization in the optimization ($/kWh). Defaults to None.
            plant_kw (float): The rating of the entire wind plant (kW).
                Defaults to None.
            x0 (iterable, optional): The initial turbine locations,
                ordered by x-coordinate and then y-coordiante
                (ie. [x1, x2, ..., xn, y1, y2, ..., yn]), and the
                initial turbine hub height (m). If none are provided, x0
                initializes to the current turbine locations and hub height.
                Defaults to None.
            bnds (iterable, optional): Bounds for the optimization
                variables (TODO: just coordinates, or height too?) (pairs of
                min/max values for each variable (m)). If none are specified,
                they are set to the min. and max. of the boundaries iterable.
                Defaults to None.
            min_dist (float, optional): The minimum distance to be
                maintained between turbines during the optimization (m). If not
                specified, initializes to 2 rotor diameters. Defaults to None.
            opt_method (str, optional): The optimization method used by
                scipy.optimize.minize. Defaults to 'SLSQP'.
            opt_options (dict, optional): Optimization options used by
                scipy.optimize.minize. If none are specified, they are set to
                {'maxiter': 100, 'disp': True, 'iprint': 2, 'ftol': 1e-9}.
                Defaults to None.
        """
        LayoutOptimization.reinitialize_opt(
            self,
            boundaries=boundaries,
            wd=wd,
            ws=ws,
            freq=freq,
            AEP_initial=AEP_initial,
            x0=x0,
            bnds=bnds,
            min_dist=min_dist,
            opt_method=opt_method,
            opt_options=opt_options,
        )

        if height_lims is not None:
            self.bndh_min = height_lims[0]
            self.bndh_max = height_lims[1]
        if COE_initial is not None:
            self.COE_initial = COE_initial
        if plant_kw is not None:
            self.plant_kw = plant_kw

        self._set_initial_conditions()
        self._set_opt_bounds_height()

    def optimize(self):
        """
        This method finds the optimized layout of wind turbines and wind
        turbine height for power production and cost of energy given the
        provided frequencies of occurance of wind conditions (wind speed,
        direction).

        Returns:
            (iterable): A list containing the optimized (x, y) locations of
            each turbine followed by the optimized height for all turbines (m).
        """
        print("=====================================================")
        print("Optimizing turbine layout and height...")
        print("Number of parameters to optimize = ", len(self.x0))
        print("=====================================================")

        opt_results_norm = self._optimize()

        print("Optimization complete.")

        opt_locs = [
            [
                self._unnorm(valx, self.bndx_min, self.bndx_max)
                for valx in opt_results_norm[0 : self.nturbs]
            ],
            [
                self._unnorm(valy, self.bndy_min, self.bndy_max)
                for valy in opt_results_norm[self.nturbs : 2 * self.nturbs]
            ],
        ]

        opt_height = [self._unnorm(opt_results_norm[-1], self.bndh_min, self.bndh_max)]

        return [opt_locs, opt_height]

    def get_farm_COE(self):
        """
        This method returns the cost of energy (COE) for the wind farm.

        Returns:
            float: The cost of energy for a wind plant in units of $/kWh.
        """
        AEP_sum = self._AEP_loop_wd()
        height = self.fi.floris.farm.turbines[0].hub_height
        return self.COE_model.COE(height, AEP_sum)



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

from typing import Any, Dict

from attrs import define, field
import numexpr as ne
import numpy as np

from floris.simulation import BaseModel
from floris.simulation import Farm
from floris.simulation import FlowField
from floris.simulation import Grid
from floris.simulation import Turbine
from floris.utilities import cosd, sind


@define
class JimenezVelocityDeflection(BaseModel):
    """
    Jiménez wake deflection model, dervied from
    :cite:`jdm-jimenez2010application`.

    References:
        .. bibliography:: /source/zrefs.bib
            :style: unsrt
            :filter: docname in docnames
            :keyprefix: jdm-
    """

    kd: float = field(default=0.05)
    ad: float = field(default=0.0)
    bd: float = field(default=0.0)
    model_string = "jimenez"

    def prepare_function(
        self,
        grid: Grid,
        flow_field: FlowField,
        turbine: Turbine
    ) -> Dict[str, Any]:

        kwargs = dict(
            x=grid.x,
            reference_rotor_diameter=turbine.rotor_diameter,
        )
        return kwargs

    # @profile
    def function(
        self,
        x_i: np.ndarray,
        y_i: np.ndarray,
        yaw_i: np.ndarray,
        turbulence_intensity_i: np.ndarray,
        ct_i: np.ndarray,
        *,
        x: np.ndarray,
        reference_rotor_diameter: np.ndarray,
    ):
        """
        Calcualtes the deflection field of the wake in relation to the yaw of
        the turbine. This is coded as defined in [1].

        Args:
            x_locations (np.array): streamwise locations in wake
            y_locations (np.array): spanwise locations in wake
            z_locations (np.array): vertical locations in wake
                (not used in Jiménez)
            turbine (:py:class:`floris.simulation.turbine.Turbine`):
                Turbine object
            coord
                (:py:meth:`floris.simulation.turbine_map.TurbineMap.coords`):
                Spatial coordinates of wind turbine.
            flow_field
                (:py:class:`floris.simulation.flow_field.FlowField`):
                Flow field object.

        Returns:
            deflection (np.array): Deflected wake centerline.


        This function calculates the deflection of the entire flow field
        given the yaw angle and Ct of the current turbine
        """

        # NOTE: Its important to remember the rules of broadcasting here.
        # An operation between two np.arrays of different sizes involves
        # broadcasting. First, the rank and then the dimensions are compared.
        # If the ranks are different, new dimensions of size 1 are added to
        # the missing dimensions. Then, arrays can be combined (arithmetic)
        # if corresponding dimensions are either the same size or 1.
        # https://numpy.org/doc/stable/user/basics.broadcasting.html
        # Here, many dimensions are 1, but these are essentially treated
        # as a scalar value for that dimension.

        # angle of deflection
        xi_init = cosd(yaw_i) * sind(yaw_i) * ct_i / 2.0

        """
        delta_x = x - x_i

        # yaw displacement
        A = 15 * (2 * self.kd * delta_x / reference_rotor_diameter + 1) ** 4.0 + xi_init ** 2.0
        B = (30 * self.kd / reference_rotor_diameter) * ( 2 * self.kd * delta_x / reference_rotor_diameter + 1 ) ** 5.0
        C = xi_init * reference_rotor_diameter * (15 + xi_init ** 2.0)
        D = 30 * self.kd

        yYaw_init = (xi_init * A / B) - (C / D)

        # corrected yaw displacement with lateral offset
        # This has the same shape as the grid

        deflection = yYaw_init + self.ad + self.bd * delta_x
        """

        # Numexpr - do not change below without corresponding changes above.
        kd = self.kd
        ad = self.ad
        bd = self.bd

        delta_x = ne.evaluate("x - x_i")
        A = ne.evaluate("15 * (2 * kd * delta_x / reference_rotor_diameter + 1) ** 4.0 + xi_init ** 2.0")
        B = ne.evaluate("(30 * kd / reference_rotor_diameter) * ( 2 * kd * delta_x / reference_rotor_diameter + 1 ) ** 5.0")
        C = ne.evaluate("xi_init * reference_rotor_diameter * (15 + xi_init ** 2.0)")
        D = ne.evaluate("30 * kd")

        yYaw_init = ne.evaluate("(xi_init * A / B) - (C / D)")
        deflection = ne.evaluate("yYaw_init + ad + bd * delta_x")

        return deflection



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

from typing import Any, Dict

from attrs import define, field
import numpy as np

from floris.simulation import BaseModel
from floris.simulation import Farm
from floris.simulation import FlowField
from floris.simulation import Grid
from floris.simulation import Turbine
from floris.utilities import cosd, sind, tand


@define
class GaussVelocityDeflection(BaseModel):
    """
    The Gauss deflection model is a blend of the models described in
    :cite:`gdm-bastankhah2016experimental` and :cite:`gdm-King2019Controls` for
    calculating the deflection field in turbine wakes.

    parameter_dictionary (dict): Model-specific parameters.
        Default values are used when a parameter is not included
        in `parameter_dictionary`. Possible key-value pairs include:

            -   **ka** (*float*): Parameter used to determine the linear
                relationship between the turbulence intensity and the
                width of the Gaussian wake shape.
            -   **kb** (*float*): Parameter used to determine the linear
                relationship between the turbulence intensity and the
                width of the Gaussian wake shape.
            -   **alpha** (*float*): Parameter that determines the
                dependence of the downstream boundary between the near
                wake and far wake region on the turbulence intensity.
            -   **beta** (*float*): Parameter that determines the
                dependence of the downstream boundary between the near
                wake and far wake region on the turbine's induction
                factor.
            -   **ad** (*float*): Additional tuning parameter to modify
                the wake deflection with a lateral offset.
                Defaults to 0.
            -   **bd** (*float*): Additional tuning parameter to modify
                the wake deflection with a lateral offset.
                Defaults to 0.
            -   **dm** (*float*): Additional tuning parameter to scale
                the amount of wake deflection. Defaults to 1.0
            -   **use_secondary_steering** (*bool*): Flag to use
                secondary steering on the wake velocity using methods
                developed in [2].
            -   **eps_gain** (*float*): Tuning value for calculating
                the V- and W-component velocities using methods
                developed in [7].
                TODO: Believe this should be removed, need to verify.
                See property on super-class for more details.

    References:
        .. bibliography:: /source/zrefs.bib
            :style: unsrt
            :filter: docname in docnames
            :keyprefix: gdm-
    """
    ad: float = field(converter=float, default=0.0)
    bd: float = field(converter=float, default=0.0)
    alpha: float = field(converter=float, default=0.58)
    beta: float = field(converter=float, default=0.077)
    ka: float = field(converter=float, default=0.38)
    kb: float = field(converter=float, default=0.004)
    dm: float = field(converter=float, default=1.0)
    eps_gain: float = field(converter=float, default=0.2)
    use_secondary_steering: bool = field(converter=bool, default=True)
    model_string = "gauss"

    def prepare_function(
        self,
        grid: Grid,
        flow_field: FlowField,
        turbine: Turbine
    ) -> Dict[str, Any]:

        kwargs = dict(
            x=grid.x,
            y=grid.y,
            z=grid.z,
            freestream_velocity=flow_field.u_initial,
            wind_veer=flow_field.wind_veer,
            reference_rotor_diameter=turbine.rotor_diameter,
        )
        return kwargs

    # @profile
    def function(
        self,
        x_i: np.ndarray,
        y_i: np.ndarray,
        yaw_i: np.ndarray,
        turbulence_intensity_i: np.ndarray,
        ct_i: np.ndarray,
        *,
        x: np.ndarray,
        y: np.ndarray,
        z: np.ndarray,
        freestream_velocity: np.ndarray,
        wind_veer: float,
        reference_rotor_diameter: float,
    ):
        """
        Calculates the deflection field of the wake. See
        :cite:`gdm-bastankhah2016experimental` and :cite:`gdm-King2019Controls`
        for details on the methods used.

        Args:
            x_locations (np.array): An array of floats that contains the
                streamwise direction grid coordinates of the flow field
                domain (m).
            y_locations (np.array): An array of floats that contains the grid
                coordinates of the flow field domain in the direction normal to
                x and parallel to the ground (m).
            z_locations (np.array): An array of floats that contains the grid
                coordinates of the flow field domain in the vertical
                direction (m).
            turbine (:py:obj:`floris.simulation.turbine`): Object that
                represents the turbine creating the wake.
            coord (:py:obj:`floris.utilities.Vec3`): Object containing
                the coordinate of the turbine creating the wake (m).
            flow_field (:py:class:`floris.simulation.flow_field`): Object
                containing the flow field information for the wind farm.

        Returns:
            np.array: Deflection field for the wake.
        """
        # ==============================================================

        # Opposite sign convention in this model
        yaw_i = -1 * yaw_i

        # TODO: connect support for tilt
        tilt = 0.0 #turbine.tilt_angle

        # initial velocity deficits
        uR = (
            freestream_velocity
          * ct_i
          * cosd(tilt)
          * cosd(yaw_i)
          / (2.0 * (1 - np.sqrt(1 - (ct_i * cosd(tilt) * cosd(yaw_i)))))
        )
        u0 = freestream_velocity * np.sqrt(1 - ct_i)

        # length of near wake
        x0 = (
            reference_rotor_diameter
            * (cosd(yaw_i) * (1 + np.sqrt(1 - ct_i * cosd(yaw_i))))
            / (np.sqrt(2) * (4 * self.alpha * turbulence_intensity_i + 2 * self.beta * (1 - np.sqrt(1 - ct_i))))
            + x_i
        )

        # wake expansion parameters
        ky = self.ka * turbulence_intensity_i + self.kb
        kz = self.ka * turbulence_intensity_i + self.kb

        C0 = 1 - u0 / freestream_velocity
        M0 = C0 * (2 - C0)
        E0 = C0 ** 2 - 3 * np.exp(1.0 / 12.0) * C0 + 3 * np.exp(1.0 / 3.0)

        # initial Gaussian wake expansion
        sigma_z0 = reference_rotor_diameter * 0.5 * np.sqrt(uR / (freestream_velocity + u0))
        sigma_y0 = sigma_z0 * cosd(yaw_i) * cosd(wind_veer)

        yR = y - y_i
        xR = x_i # yR * tand(yaw) + x_i

        # yaw parameters (skew angle and distance from centerline)
        # skew angle in radians
        theta_c0 = self.dm * (0.3 * np.radians(yaw_i) / cosd(yaw_i)) * (1 - np.sqrt(1 - ct_i * cosd(yaw_i)))
        delta0 = np.tan(theta_c0) * (x0 - x_i)  # initial wake deflection;
        # NOTE: use np.tan here since theta_c0 is radians

        # deflection in the near wake
        delta_near_wake = ((x - xR) / (x0 - xR)) * delta0 + (self.ad + self.bd * (x - x_i))
        delta_near_wake = delta_near_wake * np.array(x >= xR)
        delta_near_wake = delta_near_wake * np.array(x <= x0)

        # deflection in the far wake
        sigma_y = ky * (x - x0) + sigma_y0
        sigma_z = kz * (x - x0) + sigma_z0
        sigma_y = sigma_y * np.array(x >= x0) + sigma_y0 * np.array(x < x0)
        sigma_z = sigma_z * np.array(x >= x0) + sigma_z0 * np.array(x < x0)

        ln_deltaNum = (1.6 + np.sqrt(M0)) * (
            1.6 * np.sqrt(sigma_y * sigma_z / (sigma_y0 * sigma_z0)) - np.sqrt(M0)
        )
        ln_deltaDen = (1.6 - np.sqrt(M0)) * (
            1.6 * np.sqrt(sigma_y * sigma_z / (sigma_y0 * sigma_z0)) + np.sqrt(M0)
        )

        delta_far_wake = (
            delta0
          + theta_c0 * E0 / 5.2
          * np.sqrt(sigma_y0 * sigma_z0 / (ky * kz * M0))
          * np.log(ln_deltaNum / ln_deltaDen)
          + (self.ad + self.bd * (x - x_i))
        )

        delta_far_wake = delta_far_wake * np.array(x > x0)
        deflection = delta_near_wake + delta_far_wake

        return deflection


## GCH components

def gamma(
    D,
    velocity,
    Uinf,
    Ct,
    scale=1.0,
):
    """
    Vortex circulation strength. Units of XXX TODO

    Args:
        D (float): Rotor diameter of the current turbine
        velocity (np.array(float)): Velocities at the current turbine
        Uinf (float): Free-stream velocity
        Ct (float): Thrust coefficient at the current turbine

    Returns:
        [type]: [description]
    """
    return scale * (np.pi / 8) * D * velocity * Uinf * Ct # * cosd(yaw)  <- the cos is included in Ct


# def calculate_effective_yaw(
def wake_added_yaw(
    u_i,
    v_i,
    u_initial,
    delta_y,
    z_i,
    rotor_diameter,
    hub_height,
    ct_i,
    tip_speed_ratio,
    axial_induction_i,
    scale=1.0,
):
    """
    what yaw angle would have produced that same average spanwise velocity

    These calculations focus around the current turbine. The formulation could
    remove the dimension for n-turbines, but for consistency with other
    similar equations it is left. However, the turbine dimension should
    always have length 1.
    """

    # turbine parameters
    D = rotor_diameter              # scalar
    HH = hub_height                 # scalar
    Ct = ct_i                       # (wd, ws, 1, 1, 1) for the current turbine
    TSR = tip_speed_ratio           # scalar
    aI = axial_induction_i          # (wd, ws, 1, 1, 1) for the current turbine
    avg_v = np.mean(v_i, axis=(3,4))  # (wd, ws, 1, grid, grid)

    # flow parameters
    Uinf = np.mean(u_initial, axis=(2,3,4))
    Uinf = Uinf[:,:,None,None,None]

    # TODO: Allow user input for eps gain
    eps_gain = 0.2
    eps = eps_gain * D  # Use set value

    vel_top = ((HH + D / 2) / HH) ** 0.12 * np.ones((1, 1, 1, 1, 1))
    Gamma_top = gamma(
        D,
        vel_top,
        Uinf,
        Ct,
        scale,
    )

    vel_bottom = ((HH - D / 2) / HH) ** 0.12 * np.ones((1, 1, 1, 1, 1))
    Gamma_bottom = -1 * gamma(
        D,
        vel_bottom,
        Uinf,
        Ct,
        scale,
    )

    turbine_average_velocity = np.cbrt(np.mean(u_i ** 3, axis=(3,4)))
    turbine_average_velocity = turbine_average_velocity[:,:,:,None,None]
    Gamma_wake_rotation = 0.25 * 2 * np.pi * D * (aI - aI ** 2) * turbine_average_velocity / TSR

    ### compute the spanwise and vertical velocities induced by yaw

    # decay = eps ** 2 / (4 * nu * delta_x / Uinf + eps ** 2)   # This is the decay downstream
    yLocs = delta_y + BaseModel.NUM_EPS

    # top vortex
    # NOTE: this is the top of the grid, not the top of the rotor
    zT = z_i - (HH + D / 2) + BaseModel.NUM_EPS  # distance from the top of the grid
    rT = yLocs ** 2 + zT ** 2  # TODO: This is - in the paper
    core_shape = 1 - np.exp(-rT / (eps ** 2))  # This looks like spanwise decay - it defines the vortex profile in the spanwise directions
    v_top = (Gamma_top * zT) / (2 * np.pi * rT) * core_shape
    v_top = np.mean( v_top, axis=(3,4) )
    # w_top = (-1 * Gamma_top * yLocs) / (2 * np.pi * rT) * core_shape * decay

    # bottom vortex
    zB = z_i - (HH - D / 2) + BaseModel.NUM_EPS
    rB = yLocs ** 2 + zB ** 2
    core_shape = 1 - np.exp(-rB / (eps ** 2))
    v_bottom = (Gamma_bottom * zB) / (2 * np.pi * rB) * core_shape
    v_bottom = np.mean( v_bottom, axis=(3,4) )
    # w_bottom = (-1 * Gamma_bottom * yLocs) / (2 * np.pi * rB) * core_shape * decay

    # wake rotation vortex
    zC = z_i - HH + BaseModel.NUM_EPS
    rC = yLocs ** 2 + zC ** 2
    core_shape = 1 - np.exp(-rC / (eps ** 2))
    v_core = (Gamma_wake_rotation * zC) / (2 * np.pi * rC) * core_shape
    v_core = np.mean( v_core, axis=(3,4) )
    # w_core = (-1 * Gamma_wake_rotation * yLocs) / (2 * np.pi * rC) * core_shape * decay

    # Cap the effective yaw values between -45 and 45 degrees
    val = 2 * (avg_v - v_core) / (v_top + v_bottom)
    val = np.where(val < -1.0, -1.0, val)
    val = np.where(val > 1.0, 1.0, val)
    y = np.degrees( 0.5 * np.arcsin( val ) )

    return y[:,:,:,None,None]


def calculate_transverse_velocity(
    u_i,
    u_initial,
    delta_x,
    delta_y,
    z,
    rotor_diameter,
    hub_height,
    yaw,
    ct_i,
    tsr_i,
    axial_induction_i,
    scale=1.0
):
    """
    Calculate transverse velocity components for all downstream turbines
    given the vortices at the current turbine.
    """

    # turbine parameters
    D = rotor_diameter
    HH = hub_height
    Ct = ct_i
    TSR = tsr_i
    aI = axial_induction_i

    # flow parameters
    Uinf = np.mean(u_initial, axis=(2,3,4))
    Uinf = Uinf[:,:,None,None,None]

    eps_gain = 0.2
    eps = eps_gain * D  # Use set value

    # TODO: wind sheer is hard-coded here but should be connected to the input
    vel_top = ((HH + D / 2) / HH) ** 0.12 * np.ones((1, 1, 1, 1, 1))
    Gamma_top = sind(yaw) * cosd(yaw) * gamma(
        D,
        vel_top,
        Uinf,
        Ct,
        scale,
    )

    vel_bottom = ((HH - D / 2) / HH) ** 0.12 * np.ones((1, 1, 1, 1, 1))
    Gamma_bottom = -1 * sind(yaw) * cosd(yaw) * gamma(
        D,
        vel_bottom,
        Uinf,
        Ct,
        scale,
    )

    turbine_average_velocity = np.cbrt(np.mean(u_i ** 3, axis=(3,4)))
    turbine_average_velocity = turbine_average_velocity[:,:,:,None,None]
    Gamma_wake_rotation = 0.25 * 2 * np.pi * D * (aI - aI ** 2) * turbine_average_velocity / TSR

    ### compute the spanwise and vertical velocities induced by yaw

    # decay the vortices as they move downstream - using mixing length
    lmda = D / 8
    kappa = 0.41
    lm = kappa * z / (1 + kappa * z / lmda)
    # TODO: get this from the z input?
    z_basis = np.linspace(np.min(z), np.max(z), np.shape(u_initial)[4])
    dudz_initial = np.gradient(u_initial, z_basis, axis=4)
    nu = lm ** 2 * np.abs(dudz_initial)

    decay = eps ** 2 / (4 * nu * delta_x / Uinf + eps ** 2)   # This is the decay downstream
    yLocs = delta_y + BaseModel.NUM_EPS

    # top vortex
    zT = z - (HH + D / 2) + BaseModel.NUM_EPS
    rT = yLocs ** 2 + zT ** 2  # TODO: This is - in the paper
    core_shape = 1 - np.exp(-rT / (eps ** 2))  # This looks like spanwise decay - it defines the vortex profile in the spanwise directions
    V1 = (Gamma_top * zT) / (2 * np.pi * rT) * core_shape * decay
    W1 = (-1 * Gamma_top * yLocs) / (2 * np.pi * rT) * core_shape * decay

    # bottom vortex
    zB = z - (HH - D / 2) + BaseModel.NUM_EPS
    rB = yLocs ** 2 + zB ** 2
    core_shape = 1 - np.exp(-rB / (eps ** 2))
    V2 = (Gamma_bottom * zB) / (2 * np.pi * rB) * core_shape * decay
    W2 = (-1 * Gamma_bottom * yLocs) / (2 * np.pi * rB) * core_shape * decay

    # wake rotation vortex
    zC = z - HH + BaseModel.NUM_EPS
    rC = yLocs ** 2 + zC ** 2
    core_shape = 1 - np.exp(-rC / (eps ** 2))
    V5 = (Gamma_wake_rotation * zC) / (2 * np.pi * rC) * core_shape * decay
    W5 = (-1 * Gamma_wake_rotation * yLocs) / (2 * np.pi * rC) * core_shape * decay


    ### Boundary condition - ground mirror vortex

    # top vortex - ground
    zTb = z + (HH + D / 2) + BaseModel.NUM_EPS
    rTb = yLocs ** 2 + zTb ** 2
    core_shape = 1 - np.exp(-rTb / (eps ** 2))  # This looks like spanwise decay - it defines the vortex profile in the spanwise directions
    V3 = (-1 * Gamma_top * zTb) / (2 * np.pi * rTb) * core_shape * decay
    W3 = (Gamma_top * yLocs) / (2 * np.pi * rTb) * core_shape * decay

    # bottom vortex - ground
    zBb = z + (HH - D / 2) + BaseModel.NUM_EPS
    rBb = yLocs ** 2 + zBb ** 2
    core_shape = 1 - np.exp(-rBb / (eps ** 2))
    V4 = (-1 * Gamma_bottom * zBb) / (2 * np.pi * rBb) * core_shape * decay
    W4 = (Gamma_bottom * yLocs) / (2 * np.pi * rBb) * core_shape * decay

    # wake rotation vortex - ground effect
    zCb = z + HH + BaseModel.NUM_EPS
    rCb = yLocs ** 2 + zCb ** 2
    core_shape = 1 - np.exp(-rCb / (eps ** 2))
    V6 = (-1 * Gamma_wake_rotation * zCb) / (2 * np.pi * rCb) * core_shape * decay
    W6 = (Gamma_wake_rotation * yLocs) / (2 * np.pi * rCb) * core_shape * decay

    # total spanwise velocity
    V = V1 + V2 + V3 + V4 + V5 + V6
    W = W1 + W2 + W3 + W4 + W5 + W6

    # no spanwise and vertical velocity upstream of the turbine
    # V[delta_x < -1] = 0.0  # Subtract by 1 to avoid numerical issues on rotation
    # W[delta_x < -1] = 0.0  # Subtract by 1 to avoid numerical issues on rotation
    # TODO Should this be <= ? Shouldn't be adding V and W on the current turbine?
    V[delta_x < 0.0] = 0.0  # Subtract by 1 to avoid numerical issues on rotation
    W[delta_x < 0.0] = 0.0  # Subtract by 1 to avoid numerical issues on rotation

    # TODO: Why would the say W cannot be negative?
    W[W < 0] = 0

    return V, W


def yaw_added_turbulence_mixing(
    u_i,
    I_i,
    v_i,
    w_i,
    turb_v_i,
    turb_w_i
):
    # Since turbulence mixing is constant for the turbine,
    # use the left two dimensions only here and expand
    # before returning. Dimensions are (wd, ws).

    I_i = I_i[:,:,0,0,0]

    average_u_i = np.cbrt(np.mean(u_i ** 3, axis=(2,3,4)))

    # Convert ambient turbulence intensity to TKE (eq 24)
    k = (average_u_i * I_i) ** 2 / (2 / 3)

    u_term = np.sqrt(2 * k)
    v_term = np.mean(v_i + turb_v_i, axis=(2,3,4))
    w_term = np.mean(w_i + turb_w_i, axis=(2,3,4))

    # Compute the new TKE (eq 23)
    k_total = 0.5 * ( u_term ** 2 + v_term ** 2 + w_term ** 2 )

    # Convert TKE back to TI
    I_total = np.sqrt( (2 / 3) * k_total ) / average_u_i

    # Remove ambient from total TI leaving only the TI due to mixing
    I_mixing = I_total - I_i

    return I_mixing[:,:,None,None,None]


# def yaw_added_recovery_correction(
#     self, U_local, U, W, x_locations, y_locations, turbine, turbine_coord
# ):
#         """
#         This method corrects the U-component velocities when yaw added recovery
#         is enabled. For more details on how the velocities are changed, see [1].
#         # TODO add reference to 1

#         Args:
#             U_local (np.array): U-component velocities across the flow field.
#             U (np.array): U-component velocity deficits across the flow field.
#             W (np.array): W-component velocity deficits across the flow field.
#             x_locations (np.array): Streamwise locations in wake.
#             y_locations (np.array): Spanwise locations in wake.
#             turbine (:py:class:`floris.simulation.turbine.Turbine`):
#                 Turbine object.
#             turbine_coord (:py:obj:`floris.simulation.turbine_map.TurbineMap.coords`):
#                 Spatial coordinates of wind turbine.

#         Returns:
#             np.array: U-component velocity deficits across the flow field.
#         """
#         # compute the velocity without modification
#         U1 = U_local - U

#         # set dimensions
#         D = turbine.rotor_diameter
#         xLocs = x_locations - turbine_coord.x1
#         ky = self.ka * turbine.turbulence_intensity + self.kb
#         U2 = (np.mean(W) * xLocs) / ((ky * xLocs + D / 2))
#         U_total = U1 + np.nan_to_num(U2)

#         # turn it back into a deficit
#         U = U_local - U_total

#         # zero out anything before the turbine
#         U[x_locations < turbine_coord.x1] = 0

#         return U



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation

import numpy as np
import pandas as pd
import scipy.stats
import matplotlib.pyplot as plt


def _convert_to_numpy_array(series):
    """
    Convert an input series to NumPy array. Currently, this function
    checks if an object has a `values` attribute and returns that if it does.
    Otherwise, it returns the given input if that input is a `np.ndarray`.

    Args:
        series (pd.Series): Series to convert.

    Returns:
        np.array: Converted Series.
    """
    if hasattr(series, "values"):
        return series.values
    elif isinstance(series, np.ndarray):
        return series


# Define ci function
def ci(data, confidence=0.95):
    """
    Compute confidence interval.

    Args:
        data (np.array): Input data.
        confidence (float, optional): Confidence value. Defaults to 0.95.

    Returns:
        float: Computed confidence interval.
    """
    a = 1.0 * np.array(data)
    n = len(a)
    _, se = np.mean(a), scipy.stats.sem(a)
    h = se * scipy.stats.t.ppf((1 + confidence) / 2.0, n - 1)
    return h


def calculate_balanced_wake_loss(
    reference_power_baseline,
    test_power_baseline,
    wind_speed_array_baseline,
    wind_direction_array_baseline,
    reference_power_controlled,
    test_power_controlled,
    wind_speed_array_controlled,
    wind_direction_array_controlled,
):
    """
    Calculate balanced wake loss. See :cite:`erl-fleming2019continued` for more
    information.

    References:
        .. bibliography:: /source/zrefs.bib
            :style: unsrt
            :filter: docname in docnames
            :keyprefix: erl-

    Args:
        reference_power_baseline (np.array): Power of reference turbine in
            baseline conditions.
        test_power_baseline (np.array): Power of test turbine in baseline
            conditions.
        wind_speed_array_baseline (np.array): Wind speeds in baseline
            conditions.
        wind_direction_array_baseline (np.array): Wind  directions in baseline
            case.
        reference_power_controlled (np.array): Power of reference turbine in
            controlled conditions.
        test_power_controlled (np.array): Power of test turbine in controlled
            conditions.
        wind_speed_array_controlled (np.array): Wind speeds in controlled
            conditions.
        wind_direction_array_controlled (np.array): Wind directions in
            controlled case.

    Returns:
        dataframe: Balanced wake loss result.
    """

    # Ensure that input arrays are np.ndarray
    reference_power_baseline = _convert_to_numpy_array(reference_power_baseline)
    test_power_baseline = _convert_to_numpy_array(test_power_baseline)
    wind_speed_array_baseline = _convert_to_numpy_array(wind_speed_array_baseline)
    wind_direction_array_baseline = _convert_to_numpy_array(
        wind_direction_array_baseline
    )

    reference_power_controlled = _convert_to_numpy_array(reference_power_controlled)
    test_power_controlled = _convert_to_numpy_array(test_power_controlled)
    wind_speed_array_controlled = _convert_to_numpy_array(wind_speed_array_controlled)
    wind_direction_array_controlled = _convert_to_numpy_array(
        wind_direction_array_controlled
    )

    # Construct data frame
    df_base = pd.DataFrame(
        {
            "ref_power": reference_power_baseline,
            "test_power": test_power_baseline,
            "ws": wind_speed_array_baseline,
            "wd": wind_direction_array_baseline,
        }
    )
    df_base["con"] = "base"
    df_con = pd.DataFrame(
        {
            "ref_power": reference_power_controlled,
            "test_power": test_power_controlled,
            "ws": wind_speed_array_controlled,
            "wd": wind_direction_array_controlled,
        }
    )
    df_con["con"] = "con"
    df = df_base.append(df_con)

    # Quantize wind speed and wind direction
    df["ws"] = df.ws.round().astype(int)
    df["wd"] = df.wd.round().astype(int)

    # Look at energy loss
    df["energy_loss"] = df.ref_power - df.test_power

    # Aggragate the losses
    df = df[["ws", "wd", "con", "energy_loss", "ref_power"]]
    df["count_val"] = 1
    df_group = df.groupby(["ws", "wd", "con"]).agg([np.mean, np.std, np.sum, ci])
    df_group.columns = ["%s_%s" % c for c in df_group.columns]
    df_group = df_group.unstack()
    df_group.columns = ["%s_%s" % c for c in df_group.columns]

    # Select down a little
    df_group = df_group[df_group.count_val_sum_base > 1]
    df_group = df_group[df_group.count_val_sum_con > 1]

    # Remove the multi-index and assign back to df
    df = df_group.reset_index()

    # Compute the weighted energy loss
    df["points_per_bin"] = df.count_val_sum_base + df.count_val_sum_con
    df["wt_loss_base"] = df["energy_loss_mean_base"] * df["points_per_bin"]
    df["wt_loss_con"] = df["energy_loss_mean_con"] * df["points_per_bin"]
    df["ref_en"] = (
        (df["ref_power_mean_base"] + df["ref_power_mean_con"])
        / 2.0
        * df["points_per_bin"]
    )

    # Compute sums across ws now
    df_sum = df.groupby("wd").sum().reset_index()
    return df_sum


def plot_balanced_wake_loss(
    reference_power_baseline,
    test_power_baseline,
    wind_speed_array_baseline,
    wind_direction_array_baseline,
    reference_power_controlled,
    test_power_controlled,
    wind_speed_array_controlled,
    wind_direction_array_controlled,
    axarr,
):
    """
    Plot balanced wake loss.

    Args:
        reference_power_baseline (np.array): Array of power of
            reference turbine in baseline conditions.
        test_power_baseline (np.array): Array of power of test turbine
            in baseline conditions.
        wind_speed_array_baseline (np.array): Array of wind speeds in
            baseline conditions.
        wind_direction_array_baseline (np.array): Array of wind
            directions in baseline case.
        reference_power_controlled (np.array): Array of power of
            reference turbine in controlled conditions.
        test_power_controlled (np.array): Array of power of test
            turbine in controlled conditions.
        wind_speed_array_controlled (np.array): Array of wind speeds in
            controlled conditions.
        wind_direction_array_controlled (np.array): Array of wind
            directions in controlled case.
        axarr (list of plt.ax): A list of axes to plot onto


    Returns:
        dataframe
        # TODO ^^ inaccurate.
    """

    # Get the result frame
    df_sum = calculate_balanced_wake_loss(
        reference_power_baseline,
        test_power_baseline,
        wind_speed_array_baseline,
        wind_direction_array_baseline,
        reference_power_controlled,
        test_power_controlled,
        wind_speed_array_controlled,
        wind_direction_array_controlled,
    )

    # Plot the energy wake loss
    ax = axarr[0]

    ax.plot(df_sum.wd, df_sum.wt_loss_base / df_sum.ref_en, "b", label="Baseline")
    ax.plot(df_sum.wd, df_sum.wt_loss_con / df_sum.ref_en, "g", label="Controlled")

    # to percent
    vals = ax.get_yticks()
    ax.set_yticklabels(["{:,.0%}".format(x) for x in vals])

    ax.set_ylabel("Wake Loss (%)")
    ax.set_xlabel("Wind Direction (Deg)")
    ax.grid()
    ax.legend()

    # Look percent change
    ax = axarr[1]
    ax.plot(
        df_sum.wd,
        (df_sum.wt_loss_base - df_sum.wt_loss_con) / df_sum.ref_en,
        "g",
        label="Percent Reduction",
    )

    # to percent
    vals = ax.get_yticks()
    ax.set_yticklabels(["{:,.2%}".format(x) for x in vals])

    ax.set_ylabel("Difference in wake loss (%)")
    ax.set_xlabel("Wind Direction (Deg)")
    ax.grid()
    ax.legend()


def overall_wake_loss(
    reference_power_baseline,
    test_power_baseline,
    wind_speed_array_baseline,
    wind_direction_array_baseline,
    reference_power_controlled,
    test_power_controlled,
    wind_speed_array_controlled,
    wind_direction_array_controlled,
):
    """
    Compute the overall wake losses and changes

    Args:
        reference_power_baseline (np.array): Array of power of
            reference turbine in baseline conditions.
        test_power_baseline (np.array): Array of power of test turbine
            in baseline conditions.
        wind_speed_array_baseline (np.array): Array of wind speeds in
            baseline conditions.
        wind_direction_array_baseline (np.array): Array of wind
            directions in baseline case.
        reference_power_controlled (np.array): Array of power of
            reference turbine in controlled conditions.
        test_power_controlled (np.array): Array of power of test
            turbine in controlled conditions.
        wind_speed_array_controlled (np.array): Array of wind speeds in
            controlled conditions.
        wind_direction_array_controlled (np.array): Array of wind
            directions in controlled case.


    """
    # Get the result frame
    df_sum = calculate_balanced_wake_loss(
        reference_power_baseline,
        test_power_baseline,
        wind_speed_array_baseline,
        wind_direction_array_baseline,
        reference_power_controlled,
        test_power_controlled,
        wind_speed_array_controlled,
        wind_direction_array_controlled,
    )

    print("===OVERALL RESULTS===")
    print(
        "Baseline Energy Loss:\t%d%%"
        % (100 * (df_sum["wt_loss_base"].sum() / df_sum["ref_en"].sum()))
    )
    print(
        "Controlled Energy Loss:\t%d%%"
        % (100 * (df_sum["wt_loss_con"].sum() / df_sum["ref_en"].sum()))
    )
    print(
        "Reduction Energy Loss:\t%d%%"
        % (
            100
            * (
                df_sum["energy_loss_mean_base"].sum()
                - df_sum["energy_loss_mean_con"].sum()
            )
            / df_sum["energy_loss_mean_base"].sum()
        )
    )



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

from typing import Any, Dict

from attrs import define, field
import numexpr as ne
import numpy as np
from numpy import newaxis as na
from scipy.special import gamma

from floris.simulation import BaseModel
from floris.simulation import Farm
from floris.simulation import FlowField
from floris.simulation import Grid
from floris.simulation import Turbine
from floris.utilities import cosd, sind, tand, pshape


@define
class CumulativeGaussCurlVelocityDeficit(BaseModel):

    a_s: float = field(default=0.179367259)
    b_s: float = field(default=0.0118889215)
    c_s1: float = field(default=0.0563691592)
    c_s2: float = field(default=0.13290157)
    a_f: float = field(default=3.11)
    b_f: float = field(default=-0.68)
    c_f: float = field(default=2.41)
    alpha_mod: float = field(default=1.0)
    
    model_string = "cumulative_gauss_curl"

    def prepare_function(
        self,
        grid: Grid,
        flow_field: FlowField,
        turbine: Turbine
    ) -> Dict[str, Any]:

        kwargs = dict(
            x=grid.x,
            y=grid.y,
            z=grid.z,
            turbine_diameter=turbine.rotor_diameter,
            u_initial=flow_field.u_initial,
        )
        return kwargs

    def function(
        self,
        ii: int,
        x_i: np.ndarray,
        y_i: np.ndarray,
        z_i: np.ndarray,
        u_i: np.ndarray,
        deflection_field: np.ndarray,
        yaw_i: np.ndarray,
        turbulence_intensity: np.ndarray,
        ct: np.ndarray,
        turb_u_wake: np.ndarray,
        Ctmp: np.ndarray,
        # enforces the use of the below as keyword arguments and adherence to the
        # unpacking of the results from prepare_function()
        *,
        x: np.ndarray,
        y: np.ndarray,
        z: np.ndarray,
        turbine_diameter: np.ndarray,
        u_initial: np.ndarray,
    ) -> None:

        # print("u_i", u_i)
        # print("yaw_i", yaw_i)
        # print("turbulence_intensity_i", turbulence_intensity_i)
        # print("ct_i", ct_i)
        # print("turb_u_wake", turb_u_wake)
        # print("sigma_i", sigma_i)
        # print("Ctmp", Ctmp)

        turbine_Ct = ct
        turbine_ti = turbulence_intensity
        turbine_yaw = yaw_i

        # TODO Should this be cbrt? This is done to match v2
        turb_avg_vels = np.cbrt(np.mean(u_i ** 3, axis=(3,4)))
        turb_avg_vels = turb_avg_vels[:,:,:,None,None]

        delta_x = x - x_i

        sigma_n = wake_expansion(
            delta_x,
            turbine_Ct[:,:,ii:ii+1],
            turbine_ti[:,:,ii:ii+1],
            turbine_diameter,
            self.a_s,
            self.b_s,
            self.c_s1,
            self.c_s2,
        )

        x_i_loc = np.mean(x_i, axis=(3,4))
        x_i_loc = x_i_loc[:,:,:,None,None]

        y_i_loc = np.mean(y_i, axis=(3,4))
        y_i_loc = y_i_loc[:,:,:,None,None]

        z_i_loc = np.mean(z_i, axis=(3,4))
        z_i_loc = z_i_loc[:,:,:,None,None]

        x_coord = np.mean(x, axis=(3,4))[:,:,:,None,None]

        y_loc = y
        y_coord = np.mean(y, axis=(3,4))[:,:,:,None,None]

        z_loc = z # np.mean(z, axis=(3,4))
        z_coord = np.mean(z, axis=(3,4))[:,:,:,None,None]

        sum_lbda = np.zeros_like(u_initial)

        for m in range(0, ii - 1):
            y_coord_m = y_coord[:,:,m:m+1]
            z_coord_m = z_coord[:,:,m:m+1]

            delta_x_m = x - x_coord[:,:,m:m+1]

            sigma_i = wake_expansion(
                delta_x_m,
                turbine_Ct[:,:,m:m+1],
                turbine_ti[:,:,m:m+1],
                turbine_diameter,
                self.a_s,
                self.b_s,
                self.c_s1,
                self.c_s2,
            )

            S_i = sigma_n ** 2 + sigma_i ** 2

            Y_i = (y_i_loc - y_coord_m - deflection_field) ** 2 / (2 * S_i)
            Z_i = (z_i_loc - z_coord_m) ** 2 / (2 * S_i)

            lbda = 1.0 * sigma_i ** 2 / S_i * np.exp(-Y_i) * np.exp(-Z_i)

            sum_lbda = sum_lbda + lbda * (Ctmp[m] / u_initial)

        # Vectorized version of sum_lbda calc; has issues with y_coord (needs to be
        # down-selected appropriately. Prelim. timings show vectorized form takes
        # longer than for loop.)
        # if ii >= 2:
        #     S = sigma_n ** 2 + sigma_i[0:ii-1, :, :, :, :, :] ** 2
        #     Y = (y_i_loc - y_coord - deflection_field) ** 2 / (2 * S)
        #     Z = (z_i_loc - z_coord) ** 2 / (2 * S)

        #     lbda = self.alpha_mod * sigma_i[0:ii-1, :, :, :, :, :] ** 2 / S * np.exp(-Y) * np.exp(-Z)
        #     sum_lbda = np.sum(lbda * (Ctmp[0:ii-1, :, :, :, :, :] / u_initial), axis=0)       
        # else:
        #     sum_lbda = 0.0

        # sigma_i[ii] = sigma_n

        # blondel
        # super gaussian
        # b_f = self.b_f1 * np.exp(self.b_f2 * TI) + self.b_f3
        x_tilde = np.abs(delta_x) / turbine_diameter
        r_tilde = np.sqrt((y_loc - y_i_loc - deflection_field) ** 2 + (z_loc - z_i_loc) ** 2) / turbine_diameter

        n = self.a_f * np.exp(self.b_f * x_tilde) + self.c_f
        a1 = 2 ** (2 / n - 1)
        a2 = 2 ** (4 / n - 2)

        # based on Blondel model, modified to include cumulative effects
        C = a1 - np.sqrt(
            a2
            - (
                (n * turbine_Ct[:,:,ii:ii+1])
                * cosd(turbine_yaw)
                / (
                    16.0
                    * gamma(2 / n)
                    * np.sign(sigma_n)
                    * (np.abs(sigma_n) ** (4 / n))
                    * (1 - sum_lbda) ** 2
                )
            )
        )

        C = C * (1 - sum_lbda)

        Ctmp[ii] = C

        yR = y_loc - y_i_loc
        xR = yR * tand(turbine_yaw) + x_i

        # add turbines together
        velDef = C * np.exp((-1 * r_tilde ** n) / (2 * sigma_n ** 2))

        velDef = velDef * np.array(x - xR >= 0.1)

        turb_u_wake = turb_u_wake + turb_avg_vels * velDef
        return (
            turb_u_wake,
            Ctmp,
        )


def wake_expansion(
    delta_x,
    ct_i,
    turbulence_intensity_i,
    rotor_diameter,
    a_s,
    b_s,
    c_s1,
    c_s2,
):
    # Calculate Beta (Eq 10, pp 5 of ref. [1] and table 4 of ref. [2] in docstring)
    beta = 0.5 * (1.0 + np.sqrt(1.0 - ct_i)) / np.sqrt(1.0 - ct_i)
    k = a_s * turbulence_intensity_i + b_s
    eps = (c_s1 * ct_i + c_s2) * np.sqrt(beta)

    # Calculate sigma_tilde (Eq 9, pp 5 of ref. [1] and table 4 of ref. [2] in docstring)
    x_tilde = np.abs(delta_x) / rotor_diameter
    sigma_y = k * x_tilde + eps

    # [added dimension to get upstream values, empty, wd, ws, x, y, z  ]
    # return sigma_y[na, :, :, :, :, :, :]
    # Do this ^^ in the main function

    return sigma_y
 


# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation

import os
import re

import numpy as np
import pandas as pd

from .cut_plane import CutPlane, get_plane_from_flow_data
from .flow_data import FlowData
from ..utilities import Vec3
from ..logging_manager import LoggerBase


class SowfaInterface(LoggerBase):
    """
    Object to facilitate interaction with flow data output by SOWFA.

    Returns:
        :py:class:`floris.tools.sowfa_utilities.SowfaInterface`: object
    """

    def __init__(
        self,
        case_folder,
        flow_data_sub_path="array_mean/array.mean0D_UAvg.vtk",
        setup_sub_path="setUp",
        turbine_array_sub_path="constant/turbineArrayProperties",
        turbine_sub_path="constant/turbineProperties",
        controlDict_sub_path="system/controlDict",
        turbine_output_sub_path="turbineOutput/20000",
        assumed_settling_time=None,
    ):
        """
        SowfaInterface object init method.

        Args:
            case_folder (str): path to folder containing SOWFA data
            flow_data_sub_path (str, optional): path to mean data.
                Defaults to 'array_mean/array.mean0D_UAvg.vtk'.
            setup_sub_path (str, optional): path to setup info.
                Defaults to 'setUp'.
            turbine_array_sub_path (str, optional): path to wind plant
                info. Defaults to 'constant/turbineArrayProperties'.
            turbine_sub_path (str, optional): path to wind turbine
                info. Defaults to 'constant/turbineProperties'.
            controlDict_sub_path (str, optional): path to turbine
                controls info. Defaults to 'system/controlDict'.
            turbine_output_sub_path (str, optional): path to turbine
                operational data. Defaults to 'turbineOutput/20000'.
            assumed_settling_time (float, optional): Time to account
                for startup transients in simulation. Defaults to None.
        """
        self.logger.info(case_folder)

        # Save the case_folder and sub_paths
        self.case_folder = case_folder
        self.setup_sub_path = setup_sub_path
        self.turbine_array_sub_path = turbine_array_sub_path
        self.turbine_sub_path = turbine_sub_path
        self.controlDict_sub_path = controlDict_sub_path
        self.turbine_output_sub_path = turbine_output_sub_path

        # Read in the input files

        # Get control settings from sc input file
        # TODO Assuming not dynamic and only one setting applied for each turbine
        # TODO If not using the super controller sowfa variant, need alternative

        # Get the turbine name and locations
        turbine_array_dict = read_foam_file(
            os.path.join(self.case_folder, self.turbine_array_sub_path)
        )
        self.turbine_name = turbine_array_dict["turbineType"].replace(
            '"', ""
        )  # TODO Assuming only one type
        self.layout_x, self.layout_y = get_turbine_locations(
            os.path.join(self.case_folder, self.turbine_array_sub_path)
        )

        # Save the number of turbines
        self.num_turbines = len(self.layout_x)

        # if SC input exists, use it for yaw and pitch as it will over-ride
        # if it does not exist, assume the values in turbineArray Properties
        if os.path.exists(os.path.join(self.case_folder, "SC_INPUT.txt")):
            df_SC = read_sc_input(self.case_folder)
            self.yaw_angles = df_SC.yaw.values
            self.pitch_angles = df_SC.pitch.values
        else:
            self.logger.info(
                "No SC_INPUT.txt, getting pitch and yaw " + "from turbine array props"
            )
            self.yaw_angles = get_turbine_yaw_angles(
                os.path.join(self.case_folder, self.turbine_array_sub_path)
            )
            self.pitch_angles = get_turbine_pitch_angles(
                os.path.join(self.case_folder, self.turbine_array_sub_path)
            )
            self.logger.info(self.yaw_angles)
            self.logger.info(self.pitch_angles)

        # Get the turbine rotor diameter and hub height
        turbine_dict = read_foam_file(
            os.path.join(self.case_folder, self.turbine_sub_path, self.turbine_name)
        )
        self.D = 2 * turbine_dict["TipRad"]

        # Use the setup file and control file to determine the precursor wind
        # speed and the time flow averaging begins (settling time)
        setup_dict = read_foam_file(os.path.join(self.case_folder, self.setup_sub_path))
        controlDict_dict = read_foam_file(
            os.path.join(self.case_folder, self.controlDict_sub_path)
        )
        start_run_time = controlDict_dict["startTime"]
        averaging_start_time = setup_dict["meanStartTime"]
        if assumed_settling_time is not None:
            self.logger.info(
                "Using assumed settling time of %.1f s" % assumed_settling_time
            )
            self.settling_time = assumed_settling_time
        else:
            self.settling_time = averaging_start_time - start_run_time
        self.precursor_wind_speed = setup_dict["U0Mag"]

        # Get the wind direction
        self.precursor_wind_dir = setup_dict["dir"]

        # Get the surface roughness
        self.z0 = setup_dict["z0"]

        # Read the outputs
        self.turbine_output = read_sowfa_df(
            os.path.join(self.case_folder, self.turbine_output_sub_path)
        )

        # Remove the settling time
        self.turbine_output = self.turbine_output[
            self.turbine_output.time > self.settling_time
        ]

        # Get the sim_time
        self.sim_time_length = self.turbine_output.time.max()

        # Read the flow data
        try:
            self.flow_data = self.read_flow_frame_SOWFA(
                os.path.join(case_folder, flow_data_sub_path)
            )

            # Re-set turbine positions to flow_field origin
            self.layout_x = self.layout_x - self.flow_data.origin.x1
            self.layout_y = self.layout_y - self.flow_data.origin.x2

        except FileNotFoundError:
            self.logger.info("No flow field found, setting NULL, origin at 0")
            self.flow_data = None  # TODO might need a null flow-field

        # Try to work out the precursor directory
        self.precursor_directory = "unknown"
        try:
            with open(os.path.join(case_folder, "runscript.preprocess"), "r") as fid:
                raw = fid.readlines()

            for i, line in enumerate(raw):
                if "precursorDir=" in line:
                    self.precursor_directory = os.path.basename(
                        line.replace("precursorDir=", "")
                    )

        except FileNotFoundError:
            self.logger.info("No preprocess file found")

    def __str__(self):

        self.logger.info("---------------------")
        self.logger.info("Case: %s" % self.case_folder)
        self.logger.info("==Turbine Info==")
        self.logger.info("Turbine: %s" % self.turbine_name)
        self.logger.info("Diameter: %dm" % self.D)
        self.logger.info("Num Turbines = %d" % self.num_turbines)
        self.logger.info("==Control Settings==")
        self.logger.info("Yaw Angles, [" + ", ".join(map(str, self.yaw_angles)) + "]")
        self.logger.info(
            "Pitch Angles, [" + ", ".join(map(str, self.pitch_angles)) + "]"
        )
        self.logger.info("==Inflow Info==")
        self.logger.info("U0Mag: %.2fm/s" % self.precursor_wind_speed)
        self.logger.info("dir: %.1f" % self.precursor_wind_dir)
        self.logger.info("==Timing Info==")
        self.logger.info("Settling time: %.1fs" % self.settling_time)
        self.logger.info("Simulation time: %.1fs" % self.sim_time_length)
        self.logger.info("---------------------")

        return " "

    def calculate_horizontal_plane(
        self, height, x_resolution=200, y_resolution=200, x_bounds=None, y_bounds=None
    ):
        """
        Get a horizontal cut through plane at a specific height

        Args:
            height (float): height of cut plane, defaults to hub-height
                Defaults to Hub-height.
            x1_resolution (float, optional): output array resolution.
                Defaults to 200.
            x2_resolution (float, optional): output array resolution.
                Defaults to 200.
            x1_bounds (tuple, optional): limits of output array.
                Defaults to None.
            x2_bounds (tuple, optional): limits of output array.
                Defaults to None.

        Returns:
            horplane
        """
        # Get points from flow data
        df = get_plane_from_flow_data(
            self.flow_data, normal_vector="z", x3_value=height
        )

        # Compute and return the cutplane
        return CutPlane(df)

    def calculate_cross_plane(
        self, x_loc, x_resolution=200, y_resolution=200, x_bounds=None, y_bounds=None
    ):
        """
        Get a horizontal cut through plane at a specific height

        Args:
            height (float): height of cut plane, defaults to hub-height
                Defaults to Hub-height.
            x1_resolution (float, optional): output array resolution.
                Defaults to 200.
            x2_resolution (float, optional): output array resolution.
                Defaults to 200.
            x1_bounds (tuple, optional): limits of output array.
                Defaults to None.
            x2_bounds (tuple, optional): limits of output array.
                Defaults to None.

        Returns:
            horplane
        """
        # Get the points of data in a dataframe
        df = get_plane_from_flow_data(self.flow_data, normal_vector="x", x3_value=x_loc)

        # Compute and return the cutplane
        return CutPlane(df)

    def calculate_y_plane(
        self, y_loc, x_resolution=200, y_resolution=200, x_bounds=None, y_bounds=None
    ):
        """
        Get a horizontal cut through plane at a specific height

        Args:
            height (float): height of cut plane, defaults to hub-height
                Defaults to Hub-height.
            x1_resolution (float, optional): output array resolution.
                Defaults to 200.
            x2_resolution (float, optional): output array resolution.
                Defaults to 200.
            x1_bounds (tuple, optional): limits of output array.
                Defaults to None.
            x2_bounds (tuple, optional): limits of output array.
                Defaults to None.

        Returns:
            horplane
        """
        # Get the points of data in a dataframe
        df = get_plane_from_flow_data(self.flow_data, normal_vector="y", x3_value=y_loc)

        # Compute and return the cutplane
        return CutPlane(df)

    def get_average_powers(self):
        """
        Return the average power from the simulation per turbine

        Args:


        Returns:
            pow_list (numpy array): an array of powers per turbine
        """
        pow_list = list()
        for t in range(self.num_turbines):
            df_sub = self.turbine_output[self.turbine_output.turbine == t]
            pow_list.append(df_sub.powerGenerator.mean())
        return np.array(pow_list)

    def get_time_power_t(self, t):
        """
        Return the power over time of a specific turbine t

        Args:
        t, turbine number

        Returns:
            power
        """
        return self.turbine_output[self.turbine_output.turbine == t].powerGenerator

    def get_average_thrust(self):
        """
        Return the average thrust from the simulation per turbine

        Args:


        Returns:
            pow_list (numpy array): an array of thrust per turbine
        """
        thrust_list = list()
        for t in range(self.num_turbines):
            df_sub = self.turbine_output[self.turbine_output.turbine == t]
            thrust_list.append(df_sub.thrust.mean())
        return np.array(thrust_list)

    def read_flow_frame_SOWFA(self, filename):
        """
        Read flow array output from SOWFA

        Args:
            filename (str): name of file containing flow data.

        Returns:
            FlowData (pd.DataFrame): a pandas table with the columns,
                of all relavent flow info (e.g. x, y, z, u, v, w).
        """
        # Read the dimension info from the file
        with open(filename, "r") as f:
            for _ in range(10):
                read_data = f.readline()
                if "SPACING" in read_data:
                    splitstring = read_data.rstrip().split(" ")
                    spacing = Vec3(
                        float(splitstring[1]),
                        float(splitstring[2]),
                        float(splitstring[3]),
                    )
                if "DIMENSIONS" in read_data:
                    splitstring = read_data.rstrip().split(" ")
                    dimensions = Vec3(
                        int(splitstring[1]), int(splitstring[2]), int(splitstring[3])
                    )
                if "ORIGIN" in read_data:
                    splitstring = read_data.rstrip().split(" ")
                    origin = Vec3(
                        float(splitstring[1]),
                        float(splitstring[2]),
                        float(splitstring[3]),
                    )

        # Set up x, y, z as lists
        if dimensions.x1 > 1.0:
            xRange = np.arange(0, dimensions.x1 * spacing.x1, spacing.x1)
        else:
            xRange = np.array([0.0])

        if dimensions.x2 > 1.0:
            yRange = np.arange(0, dimensions.x2 * spacing.x2, spacing.x2)
        else:
            yRange = np.array([0.0])

        if dimensions.x3 > 1.0:
            zRange = np.arange(0, dimensions.x3 * spacing.x3, spacing.x3)
        else:
            zRange = np.array([0.0])

        pts = np.array([(x, y, z) for z in zRange for y in yRange for x in xRange])

        df = pd.read_csv(
            filename, skiprows=10, sep="\t", header=None, names=["u", "v", "w"]
        )
        x = pts[:, 0]
        y = pts[:, 1]
        z = pts[:, 2]

        return FlowData(
            x, y, z, df.u.values, df.v.values, df.w.values, spacing, dimensions, origin
        )


def read_sc_input(case_folder, wind_direction=270.0):
    """
    Read the super controller (SC) input file to get the wind farm
    control settings.

    Args:
        case_folder (str): path to folder containing SC data.
        wind_direction (float, optional): Wind direction.
            Defaults to 270..

    Returns:
        df_SC (pd.DataFrame): dataframe containing SC info.
    """
    sc_file = os.path.join(case_folder, "SC_INPUT.txt")

    df_SC = pd.read_csv(sc_file, delim_whitespace=True)

    df_SC.columns = ["time", "turbine", "yaw", "pitch"]

    df_SC["yaw"] = wind_direction - df_SC.yaw

    df_SC = df_SC.set_index("turbine")

    return df_SC


def read_sowfa_df(folder_name, channels=[]):
    """
    New function to use pandas to read in files using pandas

    Args:
        folder_name (str): where to find the outputs of ALL channels,
            not really used for now, but could be a list of desired
            channels to only read.
        channels (list, optional): list of specific channels to read.
            Defaults to [].
    """
    # Get the availble outputs
    outputNames = [
        f
        for f in os.listdir(folder_name)
        if os.path.isfile(os.path.join(folder_name, f))
    ]

    # Remove the harder input files for now (undo someday)
    # hardFiles = [
    #     "Vtangential",
    #     "Cl",
    #     "Cd",
    #     "Vradial",
    #     "x",
    #     "y",
    #     "z",
    #     "alpha",
    #     "axialForce",
    # ]
    simpleFiles = [
        "nacYaw",
        "rotSpeedFiltered",
        "rotSpeed",
        "thrust",
        "torqueGen",
        "powerRotor",
        "powerGenerator",
        "torqueRotor",
        "azimuth",
        "pitch",
    ]

    # Limit to files
    if len(channels) == 0:
        outputNames = [o for o in outputNames if o in simpleFiles]
    else:
        outputNames = channels

    # Get the number of channels
    num_channels = len(outputNames)

    if num_channels == 0:
        raise ValueError("Is %s a data folder?" % folder_name)

    # Now loop through the files
    for c_idx, chan in enumerate(outputNames):

        filename = os.path.join(folder_name, chan)

        # Load the file
        df_inner = pd.read_csv(filename, sep=" ", header=None, skiprows=1)

        # Rename the columns
        df_inner.columns = ["turbine", "time", "dt", chan]

        # Drop dt
        df_inner = df_inner[["time", "turbine", chan]].set_index(["time", "turbine"])

        # On first run declare the new frame
        if c_idx == 0:
            # Declare the main data frame to return as copy
            df = df_inner.copy(deep=True)

        # On other loops just add the new frame
        else:
            df[chan] = df_inner[chan]

    # Reset the index
    df = df.reset_index()

    # Zero the time
    df["time"] = df.time - df.time.min()

    return df


def read_foam_file(filename):
    """
    Method to read scalar and boolean/string inputs from an OpenFOAM
    input file.

    Args:
        filename (str): path to file to read.

    Returns:
        data (dict): dictionary with OpenFOAM inputs
    """
    data = {}

    with open(filename, "r") as fid:
        raw = fid.readlines()

    bloc_comment_test = False
    for i, line in enumerate(raw):

        if raw[i][0:2] == "/*":
            bloc_comment_test = True

        if not bloc_comment_test:

            # Check if the string is a comment and skip line
            if raw[i].strip()[0:2] == "//" or raw[i].strip()[0:1] == "#":
                pass

            elif len(raw[i].strip()) == 0:  # Check if the string is empty and skip line
                pass

            else:
                tmp = raw[i].strip().rstrip().split()
                try:
                    data[tmp[0].replace('"', "")] = np.float(tmp[1][:-1])
                except Exception:
                    try:
                        data[tmp[0].replace('"', "")] = tmp[1][:-1]
                    except Exception:
                        next

        if raw[i][0:2] == r"\*":
            bloc_comment_test = False

    return data


def get_turbine_locations(turbine_array_file):
    """
    Extract wind turbine locations from SOWFA data.

    Args:
        turbine_array_file (str): path to file containing wind plant
            layout data.

    Returns:
        layout_x (np.array): wind plant layout coodinates (east-west).
        layout_y (np.array): wind plant layout coodinates (north-south).
    """
    x = list()
    y = list()

    with open(turbine_array_file, "r") as f:
        for line in f:
            if "baseLocation" in line:
                # Extract the coordinates
                data = re.findall(r"[-+]?\d*\.\d+|\d+", line)

                # Append the data
                x.append(float(data[0]))
                y.append(float(data[1]))

    layout_x = np.array(x)
    layout_y = np.array(y)

    return layout_x, layout_y


def get_turbine_pitch_angles(turbine_array_file):
    """
    Extract wind turbine blade pitch information from SOWFA data.

    Args:
        turbine_array_file (str): path to file containing pitch info.

    Returns:
        p (np.array): blade pitch info.
    """
    p = list()

    with open(turbine_array_file, "r") as f:
        for line in f:
            if "Pitch" in line:
                # Extract the coordinates
                data = re.findall(r"[-+]?\d*\.\d+|\d+", line)

                # Append the data
                p.append(float(data[0]))

    return np.array(p)


def get_turbine_yaw_angles(turbine_array_file, wind_direction=270.0):
    """
    Extract wind turbine yaw angle information from SOWFA data.

    Args:
        turbine_array_file (str): path to file containing yaw info.
        wind_direction (float, optional): Wind direction.
            Defaults to 270..

    Returns:
        y (np.array): wind turbine yaw info.
    """
    y = list()

    with open(turbine_array_file, "r") as f:
        for line in f:
            if "NacYaw" in line:
                # Extract the coordinates
                data = re.findall(r"[-+]?\d*\.\d+|\d+", line)

                # Append the data
                y.append(wind_direction - float(data[0]))

    return np.array(y)



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

import copy
import numpy as np
import time
import sys

from floris.simulation import Farm
from floris.simulation import Turbine
from floris.simulation import TurbineGrid, FlowFieldGrid
from floris.simulation import Ct, axial_induction
from floris.simulation import FlowField
from floris.simulation.turbine import average_velocity
from floris.simulation.wake import WakeModelManager
from floris.simulation.wake_deflection.gauss import (
    calculate_transverse_velocity,
    wake_added_yaw,
    yaw_added_turbulence_mixing
)


def calculate_area_overlap(wake_velocities, freestream_velocities, y_ngrid, z_ngrid):
    """
    compute wake overlap based on the number of points that are not freestream velocity, i.e. affected by the wake
    """
    # Count all of the rotor points with a negligible difference from freestream
    # count = np.sum(freestream_velocities - wake_velocities <= 0.05, axis=(3, 4))
    # return (y_ngrid * z_ngrid - count) / (y_ngrid * z_ngrid)
    # return 1 - count / (y_ngrid * z_ngrid)

    # Find the points on the rotor grids with a difference from freestream of greater
    # than some tolerance. These are all the points in the wake. The ratio of
    # these points to the total points is the portion of wake overlap.
    return np.sum(freestream_velocities - wake_velocities > 0.05, axis=(3, 4)) / (y_ngrid * z_ngrid)


# @profile
def sequential_solver(farm: Farm, flow_field: FlowField, turbine: Turbine, grid: TurbineGrid, model_manager: WakeModelManager) -> None:
    # Algorithm
    # For each turbine, calculate its effect on every downstream turbine.
    # For the current turbine, we are calculating the deficit that it adds to downstream turbines.
    # Integrate this into the main data structure.
    # Move on to the next turbine.

    # <<interface>>
    deflection_model_args = model_manager.deflection_model.prepare_function(grid, flow_field, turbine)
    deficit_model_args = model_manager.velocity_model.prepare_function(grid, flow_field, turbine)

    # This is u_wake
    wake_field = np.zeros_like(flow_field.u_initial)
    v_wake = np.zeros_like(flow_field.v_initial)
    w_wake = np.zeros_like(flow_field.w_initial)

    turbine_turbulence_intensity = flow_field.turbulence_intensity * np.ones((flow_field.n_wind_directions, flow_field.n_wind_speeds, farm.n_turbines, 1, 1))
    ambient_turbulence_intensity = flow_field.turbulence_intensity

    # Calculate the velocity deficit sequentially from upstream to downstream turbines
    for i in range(grid.n_turbines):

        # Get the current turbine quantities
        x_i = np.mean(grid.x[:, :, i:i+1], axis=(3, 4))
        x_i = x_i[:, :, :, None, None]
        y_i = np.mean(grid.y[:, :, i:i+1], axis=(3, 4))        
        y_i = y_i[:, :, :, None, None]
        z_i = np.mean(grid.z[:, :, i:i+1], axis=(3, 4))
        z_i = z_i[:, :, :, None, None]

        u_i = flow_field.u[:, :, i:i+1]
        v_i = flow_field.v[:, :, i:i+1]

        ct_i = Ct(
            velocities=flow_field.u,
            yaw_angle=farm.yaw_angles,
            fCt=turbine.fCt_interp,
            ix_filter=[i],
        )
        ct_i = ct_i[:, :, 0:1, None, None]  # Since we are filtering for the i'th turbine in the Ct function, get the first index here (0:1)
        axial_induction_i = axial_induction(
            velocities=flow_field.u,
            yaw_angle=farm.yaw_angles,
            fCt=turbine.fCt_interp,
            ix_filter=[i],
        )
        axial_induction_i = axial_induction_i[:, :, 0:1, None, None]    # Since we are filtering for the i'th turbine in the axial induction function, get the first index here (0:1)
        turbulence_intensity_i = turbine_turbulence_intensity[:, :, i:i+1]
        yaw_angle_i = farm.yaw_angles[:, :, i:i+1, None, None]

        effective_yaw_i = np.zeros_like(yaw_angle_i)
        effective_yaw_i += yaw_angle_i

        if model_manager.enable_secondary_steering:
            added_yaw = wake_added_yaw(
                u_i,
                v_i,
                flow_field.u_initial,
                grid.y[:, :, i:i+1] - y_i,
                grid.z[:, :, i:i+1],
                turbine.rotor_diameter,
                turbine.hub_height,
                ct_i,
                turbine.TSR,
                axial_induction_i
            )
            effective_yaw_i += added_yaw

        # Model calculations
        # NOTE: exponential
        deflection_field = model_manager.deflection_model.function(
            x_i,
            y_i,
            effective_yaw_i,
            turbulence_intensity_i,
            ct_i,
            **deflection_model_args
        )

        if model_manager.enable_transverse_velocities:
            v_wake, w_wake = calculate_transverse_velocity(
                u_i,
                flow_field.u_initial,
                grid.x - x_i,
                grid.y - y_i,
                grid.z,
                turbine.rotor_diameter,
                turbine.hub_height,
                yaw_angle_i,
                ct_i,
                turbine.TSR,
                axial_induction_i
            )

        if model_manager.enable_yaw_added_recovery:
            I_mixing = yaw_added_turbulence_mixing(
                u_i,
                turbulence_intensity_i,
                v_i,
                flow_field.w[:, :, i:i+1],
                v_wake[:, :, i:i+1],
                w_wake[:, :, i:i+1],
            )
            gch_gain = 2
            turbine_turbulence_intensity[:, :, i:i+1] = turbulence_intensity_i + gch_gain * I_mixing

        # NOTE: exponential
        velocity_deficit = model_manager.velocity_model.function(
            x_i,
            y_i,
            z_i,
            axial_induction_i,
            deflection_field,
            yaw_angle_i,
            turbulence_intensity_i,
            ct_i,
            **deficit_model_args
        )

        wake_field = model_manager.combination_model.function(
            wake_field,
            velocity_deficit * flow_field.u_initial
        )

        wake_added_turbulence_intensity = model_manager.turbulence_model.function(
            ambient_turbulence_intensity,
            grid.x,
            x_i,
            turbine.rotor_diameter,
            axial_induction_i
        )

        # Calculate wake overlap for wake-added turbulence (WAT)
        area_overlap = np.sum(velocity_deficit * flow_field.u_initial > 0.05, axis=(3, 4)) / (grid.grid_resolution * grid.grid_resolution)
        area_overlap = area_overlap[:, :, :, None, None]

        # Modify wake added turbulence by wake area overlap
        downstream_influence_length = 15 * turbine.rotor_diameter
        ti_added = (
            area_overlap
            * np.nan_to_num(wake_added_turbulence_intensity, posinf=0.0)
            * np.array(grid.x > x_i)
            * np.array(np.abs(y_i - grid.y) < 2 * turbine.rotor_diameter)
            * np.array(grid.x <= downstream_influence_length + x_i)
        )

        # Combine turbine TIs with WAT
        turbine_turbulence_intensity = np.maximum( np.sqrt( ti_added ** 2 + ambient_turbulence_intensity ** 2 ) , turbine_turbulence_intensity )

        flow_field.u = flow_field.u_initial - wake_field
        flow_field.v += v_wake
        flow_field.w += w_wake

    flow_field.turbulence_intensity_field = np.mean(turbine_turbulence_intensity, axis=(3,4))
    flow_field.turbulence_intensity_field = flow_field.turbulence_intensity_field[:,:,:,None,None]


def full_flow_sequential_solver(farm: Farm, flow_field: FlowField, turbine: Turbine, flow_field_grid: FlowFieldGrid, model_manager: WakeModelManager) -> None:

    # Get the flow quantities and turbine performance
    turbine_grid_farm = copy.deepcopy(farm)
    turbine_grid_flow_field = copy.deepcopy(flow_field)
    turbine_grid = TurbineGrid(
        turbine_coordinates=turbine_grid_farm.coordinates,
        reference_turbine_diameter=turbine.rotor_diameter,
        wind_directions=turbine_grid_flow_field.wind_directions,
        wind_speeds=turbine_grid_flow_field.wind_speeds,
        grid_resolution=5,
    )
    turbine_grid_flow_field.initialize_velocity_field(turbine_grid)
    sequential_solver(turbine_grid_farm, turbine_grid_flow_field, turbine, turbine_grid, model_manager)

    ### Referring to the quantities from above, calculate the wake in the full grid

    # Use full flow_field here to use the full grid in the wake models
    deflection_model_args = model_manager.deflection_model.prepare_function(flow_field_grid, flow_field, turbine)
    deficit_model_args = model_manager.velocity_model.prepare_function(flow_field_grid, flow_field, turbine)

    wake_field = np.zeros_like(flow_field.u_initial)
    v_wake = np.zeros_like(flow_field.v_initial)
    w_wake = np.zeros_like(flow_field.w_initial)

    # Calculate the velocity deficit sequentially from upstream to downstream turbines
    for i in range(flow_field_grid.n_turbines):

        # Get the current turbine quantities
        x_i = np.mean(turbine_grid.x[:, :, i:i+1], axis=(3, 4))
        x_i = x_i[:, :, :, None, None]
        y_i = np.mean(turbine_grid.y[:, :, i:i+1], axis=(3, 4))        
        y_i = y_i[:, :, :, None, None]
        z_i = np.mean(turbine_grid.z[:, :, i:i+1], axis=(3, 4))
        z_i = z_i[:, :, :, None, None]

        u_i = turbine_grid_flow_field.u[:, :, i:i+1]
        v_i = turbine_grid_flow_field.v[:, :, i:i+1]

        ct_i = Ct(
            velocities=turbine_grid_flow_field.u,
            yaw_angle=turbine_grid_farm.yaw_angles,
            fCt=turbine.fCt_interp,
            ix_filter=[i],
        )
        ct_i = ct_i[:, :, 0:1, None, None]  # Since we are filtering for the i'th turbine in the Ct function, get the first index here (0:1)
        axial_induction_i = axial_induction(
            velocities=turbine_grid_flow_field.u,
            yaw_angle=turbine_grid_farm.yaw_angles,
            fCt=turbine.fCt_interp,
            ix_filter=[i],
        )
        axial_induction_i = axial_induction_i[:, :, 0:1, None, None]    # Since we are filtering for the i'th turbine in the axial induction function, get the first index here (0:1)
        turbulence_intensity_i = turbine_grid_flow_field.turbulence_intensity_field[:, :, i:i+1]
        yaw_angle_i = turbine_grid_farm.yaw_angles[:, :, i:i+1, None, None]

        effective_yaw_i = np.zeros_like(yaw_angle_i)
        effective_yaw_i += yaw_angle_i

        if model_manager.enable_secondary_steering:
            added_yaw = wake_added_yaw(
                u_i,
                v_i,
                turbine_grid_flow_field.u_initial,
                turbine_grid.y[:, :, i:i+1] - y_i,
                turbine_grid.z[:, :, i:i+1],
                turbine.rotor_diameter,
                turbine.hub_height,
                ct_i,
                turbine.TSR,
                axial_induction_i
            )
            effective_yaw_i += added_yaw

        # Model calculations
        # NOTE: exponential
        deflection_field = model_manager.deflection_model.function(
            x_i,
            y_i,
            effective_yaw_i,
            turbulence_intensity_i,
            ct_i,
            **deflection_model_args
        )

        if model_manager.enable_transverse_velocities:
            v_wake, w_wake = calculate_transverse_velocity(
                u_i,
                flow_field.u_initial,
                flow_field_grid.x - x_i,
                flow_field_grid.y - y_i,
                flow_field_grid.z,
                turbine.rotor_diameter,
                turbine.hub_height,
                yaw_angle_i,
                ct_i,
                turbine.TSR,
                axial_induction_i
            )

        # NOTE: exponential
        velocity_deficit = model_manager.velocity_model.function(
            x_i,
            y_i,
            z_i,
            axial_induction_i,
            deflection_field,
            yaw_angle_i,
            turbulence_intensity_i,
            ct_i,
            **deficit_model_args
        )

        wake_field = model_manager.combination_model.function(
            wake_field,
            velocity_deficit * flow_field.u_initial
        )

        flow_field.u = flow_field.u_initial - wake_field
        flow_field.v += v_wake
        flow_field.w += w_wake


def cc_solver(farm: Farm, flow_field: FlowField, turbine: Turbine, grid: TurbineGrid, model_manager: WakeModelManager) -> None:

    # <<interface>>
    deflection_model_args = model_manager.deflection_model.prepare_function(grid, flow_field, turbine)
    deficit_model_args = model_manager.velocity_model.prepare_function(grid, flow_field, turbine)

    # This is u_wake
    v_wake = np.zeros_like(flow_field.v_initial)
    w_wake = np.zeros_like(flow_field.w_initial)
    turb_u_wake = np.zeros_like(flow_field.u_initial)
    turb_inflow_field = copy.deepcopy(flow_field.u_initial)

    turbine_turbulence_intensity = flow_field.turbulence_intensity * np.ones((flow_field.n_wind_directions, flow_field.n_wind_speeds, farm.n_turbines, 1, 1))
    ambient_turbulence_intensity = flow_field.turbulence_intensity

    shape = (farm.n_turbines,) + np.shape(flow_field.u_initial)
    Ctmp = np.zeros((shape))
    # Ctmp = np.zeros((len(x_coord), len(wd), len(ws), len(x_coord), y_ngrid, z_ngrid))
    
    sigma_i = np.zeros((shape))
    # sigma_i = np.zeros((len(x_coord), len(wd), len(ws), len(x_coord), y_ngrid, z_ngrid))

    # Calculate the velocity deficit sequentially from upstream to downstream turbines
    for i in range(grid.n_turbines):

        # Get the current turbine quantities
        x_i = np.mean(grid.x[:, :, i:i+1], axis=(3, 4))
        x_i = x_i[:, :, :, None, None]
        y_i = np.mean(grid.y[:, :, i:i+1], axis=(3, 4))        
        y_i = y_i[:, :, :, None, None]
        z_i = np.mean(grid.z[:, :, i:i+1], axis=(3, 4))
        z_i = z_i[:, :, :, None, None]

        mask2 = np.array(grid.x < x_i + 0.01) * np.array(grid.x > x_i - 0.01) * np.array(grid.y < y_i + 0.51*126.0) * np.array(grid.y > y_i - 0.51*126.0)
        # mask2 = np.logical_and(np.logical_and(np.logical_and(grid.x < x_i + 0.01, grid.x > x_i - 0.01), grid.y < y_i + 0.51*126.0), grid.y > y_i - 0.51*126.0)
        turb_inflow_field = turb_inflow_field * ~mask2 + (flow_field.u_initial - turb_u_wake) * mask2

        turb_avg_vels = average_velocity(turb_inflow_field)
        turb_Cts = Ct(
            turb_avg_vels,
            farm.yaw_angles,
            turbine.fCt_interp,
        )
        turb_Cts = turb_Cts[:, :, :, None, None]     
        turb_aIs = axial_induction(
            turb_avg_vels,
            farm.yaw_angles,
            turbine.fCt_interp,
            ix_filter=[i],
        )
        turb_aIs = turb_aIs[:, :, :, None, None]

        u_i = turb_inflow_field[:, :, i:i+1]
        v_i = flow_field.v[:, :, i:i+1]

        axial_induction_i = axial_induction(
            velocities=flow_field.u,
            yaw_angle=farm.yaw_angles,
            fCt=turbine.fCt_interp,
            ix_filter=[i],
        )

        axial_induction_i = axial_induction_i[:, :, :, None, None]

        turbulence_intensity_i = turbine_turbulence_intensity[:, :, i:i+1]
        yaw_angle_i = farm.yaw_angles[:, :, i:i+1, None, None]

        effective_yaw_i = np.zeros_like(yaw_angle_i)
        effective_yaw_i += yaw_angle_i

        if model_manager.enable_secondary_steering:
            added_yaw = wake_added_yaw(
                u_i,
                v_i,
                flow_field.u_initial,
                grid.y[:, :, i:i+1] - y_i,
                grid.z[:, :, i:i+1],
                turbine.rotor_diameter,
                turbine.hub_height,
                turb_Cts[:, :, i:i+1],
                turbine.TSR,
                axial_induction_i,
                scale=2.0,
            )
            effective_yaw_i += added_yaw

        # Model calculations
        # NOTE: exponential
        deflection_field = model_manager.deflection_model.function(
            x_i,
            y_i,
            effective_yaw_i,
            turbulence_intensity_i,
            turb_Cts[:, :, i:i+1],
            **deflection_model_args
        )

        if model_manager.enable_transverse_velocities:
            v_wake, w_wake = calculate_transverse_velocity(
                u_i,
                flow_field.u_initial,
                grid.x - x_i,
                grid.y - y_i,
                grid.z,
                turbine.rotor_diameter,
                turbine.hub_height,
                yaw_angle_i,
                turb_Cts[:, :, i:i+1],
                turbine.TSR,
                axial_induction_i,
                scale=2.0
            )

        if model_manager.enable_yaw_added_recovery:
            I_mixing = yaw_added_turbulence_mixing(
                u_i,
                turbulence_intensity_i,
                v_i,
                flow_field.w[:, :, i:i+1],
                v_wake[:, :, i:i+1],
                w_wake[:, :, i:i+1],
            )
            gch_gain = 1.0
            turbine_turbulence_intensity[:, :, i:i+1] = turbulence_intensity_i + gch_gain * I_mixing

        turb_u_wake, Ctmp = model_manager.velocity_model.function(
            i,
            x_i,
            y_i,
            z_i,
            u_i,
            deflection_field,
            yaw_angle_i,
            turbine_turbulence_intensity,
            turb_Cts,
            turb_u_wake,
            Ctmp,
            **deficit_model_args
        )

        wake_added_turbulence_intensity = model_manager.turbulence_model.function(
            ambient_turbulence_intensity,
            grid.x,
            x_i,
            turbine.rotor_diameter,
            turb_aIs
        )

        # Calculate wake overlap for wake-added turbulence (WAT)
        area_overlap = 1 - np.sum(turb_u_wake <= 0.05, axis=(3, 4)) / (grid.grid_resolution * grid.grid_resolution)
        area_overlap = area_overlap[:, :, :, None, None]

        # Modify wake added turbulence by wake area overlap
        downstream_influence_length = 15 * turbine.rotor_diameter
        ti_added = (
            area_overlap
            * np.nan_to_num(wake_added_turbulence_intensity, posinf=0.0)
            * np.array(grid.x > x_i)
            * np.array(np.abs(y_i - grid.y) < 2 * turbine.rotor_diameter)
            * np.array(grid.x <= downstream_influence_length + x_i)
        )

        # Combine turbine TIs with WAT
        turbine_turbulence_intensity = np.maximum( np.sqrt( ti_added ** 2 + ambient_turbulence_intensity ** 2 ) , turbine_turbulence_intensity )

        flow_field.v += v_wake
        flow_field.w += w_wake
    flow_field.u = turb_inflow_field

    flow_field.turbulence_intensity_field = np.mean(turbine_turbulence_intensity, axis=(3,4))
    flow_field.turbulence_intensity_field = flow_field.turbulence_intensity_field[:,:,:,None,None]


def full_flow_cc_solver(farm: Farm, flow_field: FlowField, turbine: Turbine, flow_field_grid: FlowFieldGrid, model_manager: WakeModelManager) -> None:

    # Get the flow quantities and turbine performance
    turbine_grid_farm = copy.deepcopy(farm)
    turbine_grid_flow_field = copy.deepcopy(flow_field)
    turbine_grid = TurbineGrid(
        turbine_coordinates=turbine_grid_farm.coordinates,
        reference_turbine_diameter=turbine.rotor_diameter,
        wind_directions=turbine_grid_flow_field.wind_directions,
        wind_speeds=turbine_grid_flow_field.wind_speeds,
        grid_resolution=3,
    )
    turbine_grid_flow_field.initialize_velocity_field(turbine_grid)
    cc_solver(turbine_grid_farm, turbine_grid_flow_field, turbine, turbine_grid, model_manager)

    ### Referring to the quantities from above, calculate the wake in the full grid

    # Use full flow_field here to use the full grid in the wake models
    deflection_model_args = model_manager.deflection_model.prepare_function(flow_field_grid, flow_field, turbine)
    deficit_model_args = model_manager.velocity_model.prepare_function(flow_field_grid, flow_field, turbine)

    v_wake = np.zeros_like(flow_field.v_initial)
    w_wake = np.zeros_like(flow_field.w_initial)
    turb_u_wake = np.zeros_like(flow_field.u_initial)

    shape = (farm.n_turbines,) + np.shape(flow_field.u_initial)
    Ctmp = np.zeros((shape))

    # Calculate the velocity deficit sequentially from upstream to downstream turbines
    for i in range(flow_field_grid.n_turbines):

        # Get the current turbine quantities
        x_i = np.mean(turbine_grid.x[:, :, i:i+1], axis=(3, 4))
        x_i = x_i[:, :, :, None, None]
        y_i = np.mean(turbine_grid.y[:, :, i:i+1], axis=(3, 4))        
        y_i = y_i[:, :, :, None, None]
        z_i = np.mean(turbine_grid.z[:, :, i:i+1], axis=(3, 4))
        z_i = z_i[:, :, :, None, None]

        u_i = turbine_grid_flow_field.u[:, :, i:i+1]
        v_i = turbine_grid_flow_field.v[:, :, i:i+1]

        turb_avg_vels = average_velocity(turbine_grid_flow_field.u)
        turb_Cts = Ct(
            velocities=turb_avg_vels,
            yaw_angle=turbine_grid_farm.yaw_angles,
            fCt=turbine.fCt_interp,
        )
        turb_Cts = turb_Cts[:, :, :, None, None]

        axial_induction_i = axial_induction(
            velocities=turbine_grid_flow_field.u,
            yaw_angle=turbine_grid_farm.yaw_angles,
            fCt=turbine.fCt_interp,
            ix_filter=[i],
        )
        axial_induction_i = axial_induction_i[:, :, :, None, None]

        turbulence_intensity_i = turbine_grid_flow_field.turbulence_intensity_field[:, :, i:i+1]
        yaw_angle_i = turbine_grid_farm.yaw_angles[:, :, i:i+1, None, None]

        effective_yaw_i = np.zeros_like(yaw_angle_i)
        effective_yaw_i += yaw_angle_i

        if model_manager.enable_secondary_steering:
            added_yaw = wake_added_yaw(
                u_i,
                v_i,
                turbine_grid_flow_field.u_initial,
                turbine_grid.y[:, :, i:i+1] - y_i,
                turbine_grid.z[:, :, i:i+1],
                turbine.rotor_diameter,
                turbine.hub_height,
                turb_Cts[:, :, i:i+1],
                turbine.TSR,
                axial_induction_i,
                scale=2.0
            )
            effective_yaw_i += added_yaw

        # Model calculations
        # NOTE: exponential
        deflection_field = model_manager.deflection_model.function(
            x_i,
            y_i,
            effective_yaw_i,
            turbulence_intensity_i,
            turb_Cts[:, :, i:i+1],
            **deflection_model_args
        )

        if model_manager.enable_transverse_velocities:
            v_wake, w_wake = calculate_transverse_velocity(
                u_i,
                flow_field.u_initial,
                flow_field_grid.x - x_i,
                flow_field_grid.y - y_i,
                flow_field_grid.z,
                turbine.rotor_diameter,
                turbine.hub_height,
                yaw_angle_i,
                turb_Cts[:, :, i:i+1],
                turbine.TSR,
                axial_induction_i,
                scale=2.0
            )

        # NOTE: exponential
        turb_u_wake, Ctmp = model_manager.velocity_model.function(
            i,
            x_i,
            y_i,
            z_i,
            u_i,
            deflection_field,
            yaw_angle_i,
            turbine_grid_flow_field.turbulence_intensity_field,
            turb_Cts,
            turb_u_wake,
            Ctmp,
            **deficit_model_args
        )

        flow_field.v += v_wake
        flow_field.w += w_wake
    flow_field.u = flow_field.u_initial - turb_u_wake



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


def gaussian(x, mu, sig):
    """
    Compute gaussian function, from
    https://stackoverflow.com/questions/14873203/plotting-of-1-dimensional-gaussian-distribution-function.

    Args:
        x (np.array): Input variable to Gaussian.
        mu (float): Mean value.
        sig (float): Standard deviation.

    Returns:
        np.array: The resulting Gaussian distribution.
    """
    return np.exp(-np.power(x - mu, 2.0) / (2 * np.power(sig, 2.0)))


def _convert_to_numpy_array(series):
    """
    Convert an input series to NumPy array. Currently, this function
    checks if an object has a `values` attribute and returns that if it does.
    Otherwise, it returns the given input if that input is a `np.ndarray`.

    Args:
        series (pd.Series): Series to convert.

    Returns:
        np.array: Converted Series.
    """
    if hasattr(series, "values"):
        return series.values
    elif isinstance(series, np.ndarray):
        return series


def _calculate_bootstrap_iterations(n):
    """
    Calculate number of bootstrap iterations given length.
    # TODO: What are `bootstrap iterations`?

    Args:
        n (int): Number of points.

    Returns:
        int: Number of bootstrap iterations.
    """
    maximum = 10000
    minimum = 2000
    return int(np.round(max(min(n * np.log10(n), maximum), minimum)))


def _calculate_lower_and_upper_bound(
    bootstrap_array, percentiles, central_estimate=None, method="simple_percentile"
):
    """
    Given resultant bootstrap output array, compute lower and upper bound
    of confidence interval.

    Args:
        bootstrap_array (np.array): array of bootrapped results
            percentiles (np.array): percentile values
            central_estimate (float, optional): if not using simple percentile,
            need to provide the central estimated result. Defaults to None.
        method (str, optional): method for computing bounds. Defaults to
            'simple_percentile'.

    Returns:
        float, float:

            -   lower ci bound
            -   upper ci bound
    """
    if method == "simple_percentile":
        upper, lower = np.percentile(bootstrap_array, percentiles)
    else:
        lower, upper = 2 * central_estimate - np.percentile(
            bootstrap_array, percentiles
        )
    return lower, upper


def _get_confidence_bounds(confidence):
    """

    Get the upper and lower confidence bounds given a desired confidence level.

    Args:
        confidence (float): [description]
        # TODO: ^^

    Returns:
        float, float:

            -   upper confidence bound
            -   lower confidence bound
    """
    return [50 + 0.5 * confidence, 50 - 0.5 * confidence]


def energy_ratio(ref_pow_base, test_pow_base):
    """
    Compute the balanced energy ratio for a single binned wind direction

    Single version, not comparing controller on controller off

    This function is called to compute a single balanced
    energy ratio calculation for a particular wind direction bin.  Note
    the reference turbine should not be the turbine implementing
    control, but should be an unaffected nearby turbine, or a synthetic
    power estimate from a measurement.  See :cite:`ers-fleming2019initial` and
    :cite:`ers-fleming2019continued` for more information.

    References:
        .. bibliography:: /source/zrefs.bib
            :style: unsrt
            :filter: docname in docnames
            :keyprefix: ers-

    Args:
        ref_pow_base (np.array): Array of baseline reference turbine
            power.
        test_pow_base (np.array): Array of baseline test turbine power.


    Returns:
        tuple: tuple containing:

            -   **ratio_base** (*float*): Baseline energy ratio.
            -   **counts_base** (*float*): Number of points in baseline.
    """

    if len(ref_pow_base) == 0:
        return np.nan, np.nan

    # Weighted sums
    weight_sum_ref_base = np.sum(ref_pow_base)
    weight_sum_test_base = np.sum(test_pow_base)

    if (weight_sum_ref_base == 0) or (weight_sum_test_base == 0):
        return np.nan, np.nan

    ratio_base = weight_sum_test_base / weight_sum_ref_base

    # Get the counts
    counts_base = len(ref_pow_base)

    return ratio_base, counts_base


def calculate_balanced_energy_ratio(
    reference_power_baseline,
    test_power_baseline,
    wind_direction_array_baseline,
    wind_direction_bins,
    confidence=95,
    n_boostrap=None,
    wind_direction_bin_p_overlap=None,
):
    """
    Calculate a balanced energy ratio for each wind direction bin.

    Single version, not divided into baseline and controlled

    Calculate a balanced energy ratio for each wind direction bin.  A
    reference and test turbine are provided for the ratio, as well as
    wind speed and wind directions. These data are further divided into
    baseline and controlled conditions.  The balanced energy ratio
    function is called and used to ensure a similar distribution of
    wind speeds is used in the computation, per wind direction bin, for
    baseline and controlled results.  Resulting arrays, including upper
    and lower uncertainty bounds computed through bootstrapping, are
    returned.  Note the reference turbine should not be the turbine
    implementing control, but should be an unaffected nearby turbine,
    or a synthetic power estimate from a measurement  See
    :cite:`ers-fleming2019initial` and
    :cite:`ers-fleming2019continued` for more information.

    References:
        .. bibliography:: /source/zrefs.bib
            :style: unsrt
            :filter: docname in docnames
            :keyprefix: ers-

    Args:
        reference_power_baseline (np.array): Array of power of
            reference turbine in baseline conditions.
        test_power_baseline (np.array): Array of power of test turbine
            in baseline conditions.
        wind_speed_array_baseline (np.array): Array of wind speeds in
            baseline conditions.
        wind_direction_array_baseline (np.array): Array of wind
            directions in baseline case.
        wind_direction_bins (np.array): Wind directions bins.
        confidence (int, optional): Confidence level to use.  Defaults
            to 95.
        n_boostrap (int, optional): Number of bootstaps, if none,
            _calculate_bootstrap_iterations is called.  Defaults to
            None.
        wind_direction_bin_p_overlap (np.array, optional): Percentage
            overlap between wind direction bin. Defaults to None.

    Returns:
        tuple: tuple containing:

            **ratio_array_base** (*np.array*): Baseline energy ratio at each wind direction bin.
            **lower_ratio_array_base** (*np.array*): Lower confidence bound of baseline energy ratio at each wind direction bin.
            **upper_ratio_array_base** (*np.array*): Upper confidence bound of baseline energy ratio at each wind direction bin.
            **counts_ratio_array_base** (*np.array*): Counts per wind direction bin in baseline.
    """

    # Ensure that input arrays are np.ndarray
    reference_power_baseline = _convert_to_numpy_array(reference_power_baseline)
    test_power_baseline = _convert_to_numpy_array(test_power_baseline)
    wind_direction_array_baseline = _convert_to_numpy_array(
        wind_direction_array_baseline
    )

    # Handle no overlap specificed (assume non-overlap)
    if wind_direction_bin_p_overlap is None:
        wind_direction_bin_p_overlap = 0

    # Compute binning radius (is this right?)
    wind_direction_bin_radius = (
        (1.0 + wind_direction_bin_p_overlap / 100.0)
        * (wind_direction_bins[1] - wind_direction_bins[0])
        / 2.0
    )

    ratio_array_base = np.zeros(len(wind_direction_bins)) * np.nan
    lower_ratio_array_base = np.zeros(len(wind_direction_bins)) * np.nan
    upper_ratio_array_base = np.zeros(len(wind_direction_bins)) * np.nan
    counts_ratio_array_base = np.zeros(len(wind_direction_bins)) * np.nan

    for i, wind_direction_bin in enumerate(wind_direction_bins):

        wind_dir_mask_baseline = (
            wind_direction_array_baseline
            >= wind_direction_bin - wind_direction_bin_radius
        ) & (
            wind_direction_array_baseline
            < wind_direction_bin + wind_direction_bin_radius
        )

        reference_power_baseline_wd = reference_power_baseline[wind_dir_mask_baseline]
        test_power_baseline_wd = test_power_baseline[wind_dir_mask_baseline]
        wind_dir_array_baseline_wd = wind_direction_array_baseline[
            wind_dir_mask_baseline
        ]
        baseline_weight = gaussian(
            wind_dir_array_baseline_wd,
            wind_direction_bin,
            wind_direction_bin_radius / 2.0,
        )
        baseline_weight = baseline_weight / np.sum(baseline_weight)

        if len(reference_power_baseline_wd) == 0:
            continue

        # compute the energy ratio
        # ratio_array_base[i], counts_ratio_array_base[i] = energy_ratio(reference_power_baseline_wd, test_power_baseline_wd)

        # Get the bounds through boot strapping
        # determine the number of bootstrap iterations if not given
        if n_boostrap is None:
            n_boostrap = _calculate_bootstrap_iterations(
                len(reference_power_baseline_wd)
            )

        ratio_base_bs = np.zeros(n_boostrap)
        for i_bs in range(n_boostrap):

            # random resampling w/ replacement
            # ind_bs = np.random.randint(
            #     len(reference_power_baseline_wd), size=len(reference_power_baseline_wd))
            ind_bs = np.random.choice(
                len(reference_power_baseline_wd),
                size=len(reference_power_baseline_wd),
                p=baseline_weight,
            )
            reference_power_binned_baseline = reference_power_baseline_wd[ind_bs]
            test_power_binned_baseline = test_power_baseline_wd[ind_bs]

            # compute the energy ratio
            ratio_base_bs[i_bs], _ = energy_ratio(
                reference_power_binned_baseline, test_power_binned_baseline
            )

        # Get the confidence bounds
        percentiles = _get_confidence_bounds(confidence)

        # Compute the central over from the bootstrap runs
        ratio_array_base[i] = np.mean(ratio_base_bs)

        (
            lower_ratio_array_base[i],
            upper_ratio_array_base[i],
        ) = _calculate_lower_and_upper_bound(
            ratio_base_bs,
            percentiles,
            central_estimate=ratio_array_base[i],
            method="simple_percentile",
        )

    return (
        ratio_array_base,
        lower_ratio_array_base,
        upper_ratio_array_base,
        counts_ratio_array_base,
    )


def plot_energy_ratio(
    reference_power_baseline,
    test_power_baseline,
    wind_speed_array_baseline,
    wind_direction_array_baseline,
    wind_direction_bins,
    confidence=95,
    n_boostrap=None,
    wind_direction_bin_p_overlap=None,
    ax=None,
    base_color="b",
    label_array=None,
    label_pchange=None,
    plot_simple=False,
    plot_ratio_scatter=False,
    marker_scale=1.0,
    show_count=True,
    hide_controlled_case=False,
):
    """
    Plot the single energy ratio.

    Function mainly acts as a wrapper to call
    calculate_balanced_energy_ratio and plot the results.

    Args:
        reference_power_baseline (np.array): Array of power
            of reference turbine in baseline conditions.
        test_power_baseline (np.array): Array of power of
            test turbine in baseline conditions.
        wind_speed_array_baseline (np.array): Array of wind
            speeds in baseline conditions.
        wind_direction_array_baseline (np.array): Array of
            wind directions in baseline case.
        wind_direction_bins (np.array): Wind directions bins.
        confidence (int, optional): Confidence level to use.
            Defaults to 95.
        n_boostrap (int, optional): Number of bootstaps, if
            none, _calculate_bootstrap_iterations is called.  Defaults
            to None.
        wind_direction_bin_p_overlap (np.array, optional):
            Percentage overlap between wind direction bin. Defaults to
            None.
        axarr ([axes], optional): list of axes to plot to.
            Defaults to None.
        base_color (str, optional): Color of baseline in
            plots. Defaults to 'b'.
        label_array ([str], optional): List of labels to
            apply Defaults to None.
        label_pchange ([type], optional): Label for
            percentage change. Defaults to None.
        plot_simple (bool, optional): Plot only the ratio, no
            confidence. Defaults to False.
        plot_ratio_scatter (bool, optional): Include scatter
            plot of values, sized to indicate counts. Defaults to False.
        marker_scale ([type], optional): Marker scale.
            Defaults to 1.
        show_count (bool, optional): Show the counts as scatter plot
        hide_controlled_case (bool, optional): Option to hide the control case from plots, for demonstration

    """

    if ax is None:
        fig, ax = plt.subplots(1, 1, sharex=True)

    if label_array is None:
        label_array = ["Baseline"]

    if label_pchange is None:
        label_pchange = "Energy Gain"

    (
        ratio_array_base,
        lower_ratio_array_base,
        upper_ratio_array_base,
        counts_ratio_array_base,
    ) = calculate_balanced_energy_ratio(
        reference_power_baseline,
        test_power_baseline,
        wind_direction_array_baseline,
        wind_direction_bins,
        confidence=95,
        n_boostrap=None,
        wind_direction_bin_p_overlap=wind_direction_bin_p_overlap,
    )

    if plot_simple:
        ax.plot(
            wind_direction_bins,
            ratio_array_base,
            label=label_array[0],
            color=base_color,
            ls="--",
        )
        ax.axhline(1, color="k")
        ax.set_ylabel("Energy Ratio (-)")

    else:

        ax.plot(
            wind_direction_bins,
            ratio_array_base,
            label=label_array[0],
            color=base_color,
            ls="-",
            marker=".",
        )
        ax.fill_between(
            wind_direction_bins,
            lower_ratio_array_base,
            upper_ratio_array_base,
            alpha=0.3,
            color=base_color,
            label="_nolegend_",
        )
        if show_count:
            ax.scatter(
                wind_direction_bins,
                ratio_array_base,
                s=counts_ratio_array_base,
                label="_nolegend_",
                color=base_color,
                marker="o",
                alpha=0.2,
            )
        ax.axhline(1, color="k")
        ax.set_ylabel("Energy Ratio (-)")

        ax.grid(True)
        ax.set_xlabel("Wind Direction (Deg)")



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

from attrs import define
import numpy as np

from floris.simulation import BaseModel


@define
class SOSFS(BaseModel):
    """
    SOSFS uses sum of squares freestream superposition to combine the
    wake velocity deficits to the base flow field.
    """

    model_string = "crespo_hernandez"

    def prepare_function(self) -> dict:
        pass

    def function(self, wake_field: np.ndarray, velocity_field: np.ndarray):
        """
        Combines the base flow field with the velocity defecits
        using sum of squares.

        Args:
            u_field (np.array): The base flow field.
            u_wake (np.array): The wake to apply to the base flow field.

        Returns:
            np.array: The resulting flow field after applying the wake to the
                base.
        """
        return np.hypot(wake_field, velocity_field)
